# -*- coding: utf-8 -*-
"""
This module provides 1d to nD grid utilities to get information
or perform operations on a grid.

For operations between different grids, please see :mod:`xoa.regrid`.
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import xarray as xr

from .__init__ import XoaError, xoa_warn
from . import misc
from . import cf
from . import coords as xcoords


def apply_along_dim(
    ds, dim, func, coord_func=None, data_kwargs=None, coord_kwargs=None, name_kwargs=None, **kwargs
):
    """Apply an operator on data array or dataset dimensions

    The operator may potentially change size of the array.
    It is is applied on the data array with the data_kwargs
    arguments and the coordinate arrays with the coord_kwargs arguments.

    Parameters
    ----------
    ds: xarray.DataArray, xarray.Dataset
    dim: str, tuple(str)
    func: callable
        Operator function that works on a specific dimension.
        It is applied to both data and coordinates, unless
        ``coord_func`` is provided.
    coord_func: callable, None
        Fonction to apply to coordinates specifically, which defaults
        to ``func``
    data_kwargs: None, dict
        Parameters passed to func for the data array
    coord_kwargs: None, dict
        Parameters passed to func for the coordinates
    name_kwargs: dict(dict)
        A dict of whose keys are coordinate name and whose values
        are passed to func only for these coordinates.
    kwargs: dict
        Extra keywords are passed to the ``func`` function

    Return
    ------
    xarray.DataArray, xarray.Dataset

    See also
    --------
    get_centers
    get_edges
    pad
    """
    # Always return a copy
    dso = ds.copy()

    # Loop on dims
    if coord_func is None:
        coord_func = func
    dim = cf.get_cf_specs(ds).parse_dims(dim, ds)
    dims = (dim,) if isinstance(dim, str) else dim
    for dim in dims:
        if dim not in dso.dims:
            continue

        # Data array or dataset
        old_coords = dso.coords
        if isinstance(dso, xr.Dataset):
            das = dso.data_vars.values()
            dso = xr.Dataset()
        else:
            das = [dso]
        daos = {}
        kwd = kwargs.copy()
        if data_kwargs:
            kwd.update(data_kwargs)
        for da in das:
            if dim not in da.dims:
                dao = da
            else:
                kw = kwd.copy()
                if name_kwargs and da.name in name_kwargs:
                    kw.update(name_kwargs.get(da.name))
                dao = func(xr.DataArray(da.data, dims=da.dims), dim, **kw)
                dao.name = da.name
                dao.encoding = da.encoding
                dao.attrs = da.attrs
            daos[dao.name] = dao
        if isinstance(dso, xr.Dataset):
            dso = dso.update(daos)
            dso.attrs = ds.attrs
            dso.encoding = ds.encoding
        else:
            dso = list(daos.values())[0]
        da_names = [name for name in daos.keys() if name]

        # Coordinates
        coords = {}
        if name_kwargs is None:
            name_kwargs = {}
        for coord_name, old_coord in old_coords.items():
            if coord_name in da_names:
                continue
            if dim in old_coord.dims:
                kw = kwargs or {}
                for dd in (coord_kwargs, name_kwargs.get(coord_name)):
                    if dd:
                        kw.update(dd)
                coord = coord_func(xr.DataArray(old_coord.data, dims=old_coord.dims), dim, **kw)
                coord.attrs = old_coord.attrs
                coord.encoding = old_coord.encoding
            else:
                coord = old_coord
            coords[coord_name] = coord
        dso = dso.assign_coords(coords)

    cf.assign_cf_specs(dso, ds)

    return dso


def _pad_(da, dim, pad_width, mode, **kwargs):

    pad_width = pad_width.get(dim, 0)
    if not pad_width:
        return da.copy()

    if mode != "linear_extrap":
        return da.pad({dim: pad_width}, mode=mode, **kwargs)

    to_concat = []
    if isinstance(pad_width, int):
        pad_width = (pad_width,)
    pad_width0 = pad_width[0]
    pad_width1 = pad_width[-1]
    if not pad_width0 and not pad_width1:
        return da
    if pad_width0:
        ramp0 = xr.DataArray(np.arange(pad_width0, 0, -1, dtype=da.dtype), dims=dim)
        da0 = da[{dim: 0}] + (da[{dim: 0}] - da[{dim: 1}]) * ramp0
        to_concat.append(da0.transpose(*da.dims))
    to_concat.append(da)
    if pad_width1:
        ramp1 = xr.DataArray(np.arange(1, pad_width1 + 1, dtype=da.dtype), dims=dim)
        da1 = da[{dim: -1}] + (da[{dim: -1}] - da[{dim: -2}]) * ramp1
        to_concat.append(da1.transpose(*da.dims))

    return xr.concat(to_concat, dim=dim)


def pad(da, pad_width, mode="edge", coord_mode="linear_extrap", name_kwargs=None, **kwargs):
    """Pad data and coordinates along dimensions

    This function adds the ``"linear_extrap"`` mode support to the builtin
    :meth:`xarray.DataArray.pad` methods.

    Parameters
    ----------
    da: xarray.DataArray
    pad_width: dict
        Pad widths. Keys are dimensions and values are int or tuple of ints.
    mode: str
        Extrapolation mode for the data array
    coord_mode: str
        Extrapolation mode for the coordinates
    name_kwargs: dict(dict)
        Keys are coordinates names and valkues are parameters to pass
        to :func:`xarray.pad` for this coordinate array
    kwargs:
        Extra arguments are passed to :func:`xarray.pad`

    Return
    ------
    xarray.DataArray

    See also
    --------
    get_centers
    get_edges
    apply_along_dim
    xarray.pad
    """
    pad_width = cf.get_cf_specs(da).parse_dims(pad_width, da)
    return apply_along_dim(
        da,
        list(pad_width.keys()),
        _pad_,
        data_kwargs={"mode": mode, **kwargs},
        coord_kwargs={"mode": coord_mode},
        name_kwargs=name_kwargs,
        pad_width=pad_width,
    )


def _get_centers_(da, dim):
    dao = da.isel({dim: slice(None, -1)})
    dao = dao + 0.5 * da.diff(dim).data
    return dao


def get_centers(da, dim):
    """Interpolate the data array at mid grid points along the `dim` dimension(s)

    .. note:: Coordinates are also centered

    Parameters
    ----------
    da: xarray.DataArray
    dim: str, tuple
        Single or tuple of data-array or generic dimension names.

    Return
    ------
    xarray.DataArray

    See also
    --------
    pad
    get_edges
    apply_along_dim
    """
    dim = cf.get_cf_specs(da).parse_dims(dim, da)
    return apply_along_dim(da, dim, _get_centers_)


def get_edges(da, dim, mode="edge", **kwargs):
    """Interpolate and extrapolate a data array at grid edges along the `dim` dimension(s)

    .. note:: Coordinates are linearly extrapolated

    Parameters
    ----------
    da: xarray.DataArray
    dim: str, tuple
        Single or tuple of data-array or generic dimension names.
    mode: str
        Extrapolation mode at grid edges
    kwargs:
        Extra arguments are passed to :func:`pad`

    Return
    ------
    xarray.DataArray

    See also
    --------
    pad
    get_centers
    apply_along_dim
    """
    # Extrapolate
    dim = cf.get_cf_specs(da).parse_dims(dim, da)
    dims = (dim,) if isinstance(dim, str) else dim
    pad_width = dict((dim, 1) for dim in dims)
    da = pad(da, pad_width=pad_width, mode=mode, **kwargs)

    # Inner edges
    return get_centers(da, dim)


class shift_directions(misc.IntEnumChoices, metaclass=misc.XEnumMeta):
    """Shift directions for :func:`shift`"""

    #: To the left/bottom/west/south/low
    left = -1
    #: To the left/bottom/west/south/low
    bottom = -1
    #: To the left/bottom/west/south/low
    south = -1
    #: To the left/bottom/west/south/low
    low = -1
    #: To the left/bottom/west/south/low
    west = -1
    #: To the right/top/east/north/high
    right = 1
    #: To the right/top/east/north/high
    top = 1
    #: To the right/top/east/north/high
    north = 1
    #: To the right/top/east/north/high
    high = 1
    #: To the right/top/east/north/high
    east = 1


def shift(da, shift_dirs, mode="edge", **kwargs):
    """Shift the grid by an half grid cell along specified dimensions and directions

    This is typically useful with Arakawa grids.

    Parameters
    ----------
    da: xarray.DataArray, xarray.Dataset
    shift_dirs: dict
        Keys are dimension names and values are directions:
        {shift_directions.rst_with_links}
    mode: str
        Extrapolation mode at grid edges
    kwargs:
        Extra arguments are passed to :func:`pad`

    Return
    ------
    xarray.DataArray, xarray.Dataset

    See also
    --------
    pad
    get_edges
    get_centers
    """
    shift_dirs = cf.get_cf_specs(da).parse_dims(shift_dirs, da)

    # Extrapolate
    pad_width = {}
    for dim, shift_dir in shift_dirs.items():
        pad_width[dim] = (1, 0) if shift_directions[shift_dir] < 0 else (0, 1)
    da = pad(da, pad_width=pad_width, mode=mode, **kwargs)

    # Inner edges
    return get_centers(da, list(shift_dirs.keys()))


shift.__doc__ = shift.__doc__.format(**locals())


def _diff_(da, dim):
    return da.diff(dim)


def diff(da, dim):
    """Compute the difference between consecutive grid points

    .. note:: Coordinates are centered between grid point with :func:`get_centers`

    Parameters
    ----------
    da: xarray.DataArray
    dim: str, tuple

    Return
    ------
    xarray.DataArray

    See also
    --------
    pad
    get_edges
    get_centers
    apply_along_dim
    """
    return apply_along_dim(da, dim, _diff_, coord_func=_get_centers_)


class dz2depth_ref_types(misc.IntEnumChoices, metaclass=misc.DefaultEnumMeta):
    """Integration ref types for :func:`dz2depth`"""

    #: Infer it (default)
    infer = 0
    #: Up (SSH)
    top = 1
    #: Up (SSH)
    ssh = 1
    #: Bottom (bathy)
    bottom = -1
    #: Bottom (bathy)
    bathy = -1


def dz2depth(dz, positive=None, zdim=None, ref=None, ref_type='infer', centered=False):
    """Integrate layer thicknesses to compute depths

    The output depths are the depths at the bottom of the layers and the top
    is at a depth of zero. Thus, the output array has the same dimensions
    as the input array of layer thinknesses.

    Parameters
    ----------
    dz: xarray.DataArray
        Layer thinknesses
    positive: str, int, None
        Direction over wich coordinates are increasing:
        {xcoords.positive_attr.rst_with_links}
        When "up", the first level is supposed to be the bottom
        and the output coordinates are negative.
        When "down", first level is supposed to be the top
        and the output coordinates are positive.
        When "guess", the dz array must have an axis coordinate
        of the same name as the z dimension, and this coordinate must have
        a valid positive attribute.
    zdim: str
        Name of the vertical dimension.
        If note set, it is infered with :func:`~xoa.coords.get_cf_dims`.
    ref: xarray.DataArray
        Reference array converting layer thicknesses to depth:

        - If **positive up**, it is expected to be the **SSH** (sea surface heigth)
          by default
        - If **positive down**, it is expected to be by default the depth of ground
          also known as **bathymetry**, which should be positive.

    ref_type: str, int
        Type of `ref`:
        {dz2depth_ref_types.rst_with_links}
    centered: bool
        Get depth a the middle of layers instead at their edge

    Return
    ------
    xarray.DataArray
        Output depths with the same dimensions as input array.

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.grid import dz2depth
        @suppress
        import xarray as xr
        dz = xr.DataArray([1., 3., 4.], dims="nz")

        # Positive down
        print(dz2depth(dz, "down"))

        # Positive up
        print(dz2depth(dz, "up"))
    """
    # Vertical dimension
    if zdim is None:
        zdim = xcoords.get_zdim(dz, errors="raise")

    # Positive attribute
    positive = xcoords.positive_attr[positive].name
    if positive == "infer":
        positive = xcoords.get_positive_attr(dz, zdim)
        if positive is None:
            raise XoaError("Can't infer positive attribute from data array/dataset")

    # Integrate
    depth = dz.cumsum(dim=zdim)
    depth = pad(depth, {zdim: (1, 0)}, mode="constant", constant_values=0)
    ref_type = dz2depth_ref_types[ref_type].name
    cfspecs = cf.get_cf_specs(dz)
    if ref is None and ref_type == "infer":
        if cfspecs.data_vars.match(ref, "bathy"):
            ref_type = "bottom"
        elif cfspecs.data_vars.match(ref, "ssh"):
            ref_type = "top"
        else:
            ref_type = "top" if positive == "down" else "bottom"
    if positive == "up":
        if ref is None:
            ref = depth.isel({zdim: -1})
        elif ref is not None and ref_type == "top":
            ref = depth.isel({zdim: -1}) - ref
        depth[:] -= ref
    else:
        if ref is not None:
            if ref_type == "bottom":
                depth[:] -= depth.isel({zdim: -1})
            depth[:] += ref

    # Fix index
    if zdim in depth.indexes:
        dnz = depth[zdim].diff(zdim).pad({zdim: (0, 1)}, mode="edge")
        depth = xcoords.change_index(depth, zdim, depth[zdim] + 0.5 * dnz.data)

    # Centered
    if centered:
        depth = get_centers(depth, zdim)
        if zdim in depth.indexes:
            depth = depth.assign_coords({zdim: dz[zdim]})

    # Finalize
    depth.attrs["positive"] = positive
    depth = cfspecs.format_coord(
        depth, "depth", rename=True, format_coords=False, rename_dims=False
    )

    return depth


dz2depth.__doc__ = dz2depth.__doc__.format(**locals())


@misc.ERRORS.format_function_docstring
def decode_cf_dz2depth(ds, errors="raise", **kwargs):
    """Compute depth from layer thickness in a dataset

    This makes use of the :meth:`~xoa.cf.CFSpecs` instance that is retreived
    with :func:`xoa.cf.get_cf_specs` with ds as an argument in order to
    find needed variables.

    Parameters
    ----------
    ds: xarray.Dataset
        Dataset that contains everything
    {errors}
    kwargs: dict
        Extra keywords are passed to :func:`dz2depth`

    Return
    ------
    xarray.Dataset
        A new dataset with a depth coordinate

    See also
    --------
    dz2depth
    xoa.cf.get_cf_specs
    """
    ds = ds.copy()
    errors = misc.ERRORS[errors]

    # Find needed stuff
    cfspecs = cf.get_cf_specs(ds)
    dz = cfspecs.search(ds, 'dz', errors=errors)
    if dz is None:
        return ds
    zdims = xcoords.get_cf_dims(dz, "z", errors=errors)
    if zdims is None:
        return ds
    zdim = zdims[0]
    positive = cfspecs["vertical"]["positive"]
    if positive is None:
        positive = xcoords.get_positive_attr(ds, zdim)
    if positive is None:
        msg = "Can't infer positive attribute from data dataset"
        if errors == "raise":
            raise XoaError(msg)
        xoa_warn(msg)
        return ds
    ssh = cfspecs.search(ds, 'ssh', errors="ignore")
    bathy = cfspecs.search(ds, 'bathy', errors="ignore")

    # Make choices
    if ssh is None and bathy is None:
        ref, ref_type = None, "infer"
    else:
        for ref, ref_type in [(bathy, "bathy"), (ssh, "ssh")][:: int(positive)]:
            if ref is not None:
                break

    # Compute depth
    depth = dz2depth(dz, positive=positive, zdim=zdim, ref=ref, ref_type=ref_type, centered=True)

    # Assign to dataset
    return ds.assign_coords(depth=depth)


def to_rect(da, tol=1e-5):
    """Convert a curvilinear coordinate array to a rectangular 1d coordinate array

    It checks if the coordinates may be converted to 1D without loss of information.

    Parameters
    ----------
    da: xarray.DataArray, xarray.Dataset
        In case of a dataset, it must contain longitudes and latitudes.

    Return
    ------
    xarray.DataArray, xarray.Dataset
    """
    da = da.copy()
    new_coords = {}
    rename_args = {}
    for name, coord in da.coords.items():
        if coord.ndim != 2:
            continue
        if xcoords.is_lon(coord):
            odim = xcoords.get_ydim(coord, errors="ignore")
        elif xcoords.is_lat(coord):
            odim = xcoords.get_xdim(coord, errors="ignore")
        else:
            continue
        dims = [odim] if odim else coord.dims
        for odim in dims:
            if np.allclose(coord.min(odim), coord.max(odim), atol=tol):
                new_coords[name] = xr.DataArray(
                    coord.isel({odim: 0}).values, dims=name, attrs=coord.attrs
                )
                new_coords[name].encoding.update(coord.encoding)
                dim = coord.dims[0] if coord.dims[1] == odim else coord.dims[1]
                rename_args[dim] = name
                break
    if new_coords:
        return (
            da.reset_coords(list(new_coords), drop=True)
            .rename(rename_args)
            .assign_coords(new_coords)
        )
    return da



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.filter` module
"""

import pytest
import numpy as np
import xarray as xr

from xoa import filter as xfilter

np.random.seed(0)
da1d = xr.DataArray(np.random.normal(size=10), dims=('nt', ))
da2d = xr.DataArray(np.random.normal(size=(50, 50)), dims=('ny', 'nx'))
da3d = xr.DataArray(np.random.normal(size=(30, 50, 50)),
                    dims=('nt', 'ny', 'nx'))


def test_filter_generate_orthogonal_kernel():
    kernel = xfilter.generate_orthogonal_kernel((3, 5), "bartlett")
    assert kernel.shape == (3, 5)
    np.testing.assert_allclose(kernel[1], [0., 0.5, 1., 0.5, 0.])

    kernel = xfilter.generate_orthogonal_kernel([2.8], "ones", fill_value=-1)
    np.testing.assert_allclose(kernel, [-1, 1, 1, -1])


@pytest.mark.parametrize(
    "kernel, data, oshape, odims, osum",
    [
      (3, da1d, (3, ), ('nt', ), 3),
      ((4, 3), da2d, (4, 3), ('ny', 'nx'), 12),
      ([1, 2, 1], da1d, (3,), ('nt', ), 4),
      ([1, 2, 1], da2d, (3, 3), ('ny', 'nx'), 16),
      ({'nx': [1, 2, 1], 'ny': [1, 1, 2, 1, 1]}, da2d,
       (5, 3), ('ny', 'nx'), 24),
      (np.ones((3, 5)), da2d, (3, 5), ('ny', 'nx'), 15),
      (xr.DataArray(np.ones((3, 5)), dims=('nx', 'ny')), da2d,
       (5, 3), ('ny', 'nx'), 15),
      (xr.DataArray(np.ones(3), dims='nx'), da2d, (1, 3), ('ny', 'nx'), 3),
    ]
)
def test_filter_generate_kernel(kernel, data, oshape, odims, osum):
    okernel = xfilter.generate_kernel(kernel, data)
    assert okernel.shape == oshape
    assert okernel.dims == odims
    np.testing.assert_allclose(okernel.data.sum(), osum)


def test_convolve():

    da = xr.DataArray(np.ones((5, 5)), dims=('ny', 'nx'))
    da[2, 2] = np.nan

    dac = xfilter.convolve(da, 3, normalize=True)
    np.testing.assert_allclose(dac.data, np.ones(da.shape))

    dac = xfilter.convolve(da, 3, normalize=False)
    assert dac.sum() == 160


def test_erode_mask():

    da_in = xr.DataArray(np.random.normal(size=(2, 9, 11)), dims=('nt', 'ny', 'nx'))
    nt, ny, nx = da_in.shape
    i0, n = 2, 5
    da_in[:, i0:i0+n, i0:i0+n] = np.nan

    # iteration
    da_out_iter = xfilter.erode_mask(da_in, 2)
    assert int(da_out_iter[1].count()) == (
        da_in[1].size - (n-4)**2)

    # until mask
    mask = np.isnan(da_out_iter)
    da_out_until = xfilter.erode_mask(da_in, mask)
    np.testing.assert_allclose(da_out_iter, da_out_until)


def test_erode_coast():

    da_in = xr.DataArray(np.random.normal(size=(3, 3, 3)), dims=('nt', 'ny', 'nx'))
    da_in[:, 1, 1] = np.nan

    np_out = xfilter.erode_coast(da_in, 1).values
    np_in = da_in.values
    np.testing.assert_allclose(
        np_out[1, 1, 1], 0.25*(np_in[1, 0, 1]+np_in[1, 2, 1]+np_in[1, 1, 0]+np_in[1, 1, 2]))



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.grid` module
"""
import functools
import warnings
import numpy as np
import xarray as xr
import pytest

import xoa
from xoa import grid


@functools.lru_cache()
def get_da():
    x = xr.DataArray(np.arange(4), dims='x')
    y = xr.DataArray(np.arange(3), dims='y')
    lon = xr.DataArray(np.resize(x*2., (3, 4)), dims=('y', 'x'))
    lat = xr.DataArray(np.resize(y*3., (4, 3)).T, dims=('y', 'x'))
    z = xr.DataArray(np.arange(2), dims='z')
    dep = z*(lat+lon) * 100
    da = xr.DataArray(
        np.resize(lat-lon, (2, 3, 4)),
        dims=('z', 'y', 'x'),
        coords={'dep': dep, 'lat': lat, "z": z, 'lon': lon, 'y': y, 'x': x},
        attrs={"long_name": "SST"},
        name="sst")
    da.encoding.update(cfspecs='croco')
    return da


def test_grid_get_centers():

    da = get_da()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", ".*cf.*")
        dac = grid.get_centers(da, dim=("y", "x"))
    assert dac.shape == (da.shape[0], da.shape[1]-1, da.shape[2]-1)
    assert dac.x[0] == 0.5
    assert dac.y[0] == 0.5
    assert dac.lon[0, 0] == 1.
    assert dac.lat[0, 0] == 1.5
    assert dac.dep[-1, 0, 0] == 250
    assert dac[0, 0, 0] == 0.5
    assert dac.name == 'sst'
    assert dac.long_name == "SST"
    assert dac.encoding["cfspecs"] == "croco"


def test_grid_pad():

    da = get_da()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", ".*cf.*")
        warnings.filterwarnings("error", ".*ist-or-tuple.*")
        dap = grid.pad(da, {"y": 1, "x": 1}, name_kwargs={'dep': {"mode": 'edge'}})
    assert dap.shape == (da.shape[0], da.shape[1]+2, da.shape[2]+2)
    assert dap.x[0] == -1
    assert dap.x[-1] == da.sizes['x']
    assert dap.y[0] == -1
    assert dap.y[-1] == da.sizes['y']
    assert dap.lon[0, 0] == -2
    assert dap.lat[0, 0] == -3
    assert dap.dep[-1, 0, 0] == da.dep[-1, 0, 0]
    assert dap[-1, 0, 0] == da[-1, 0, 0]
    assert dap.name == 'sst'
    assert dap.long_name == "SST"
    assert dap.encoding["cfspecs"] == "croco"


def test_grid_get_edges():

    da = get_da()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", ".*cf.*")
        dae = grid.get_edges(da, "y")
    assert dae.shape == (da.shape[0], da.shape[1]+1, da.shape[2])
    np.testing.assert_allclose(dae.y[:2].values, da.y[:2] - 0.5)
    np.testing.assert_allclose(dae.lat[:2, 0], da.lat[:2, 0] - 1.5)
    assert dae[1, 0, 0] == da[1, 0, 0]


def test_grid_shift():
    da = get_da()
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", ".*cf.*")
        dae = grid.shift(da, {"x": "left", "y": 1})
    assert dae.shape == da.shape
    np.testing.assert_allclose(dae.x.values, da.x - 0.5)
    np.testing.assert_allclose(dae.y.values, da.y + 0.5)
    assert dae[1, 0, 1] == float(da[1, :2, :2].mean())


@pytest.mark.parametrize(
    "positive, expected, ref, ref_type", [
        ["down", [0, 100., 600., 1600.], None, None],
        ["down", [10, 110., 610., 1610.], 10, "top"],
        ["down", [15, 115., 615., 1615.], 1615, "bottom"],
        ["up", [-1600, -1500, -1000, 0], None, None],
        ["up", [-1610, -1510, -1010, -10], 1610, "bottom"],
        ["up", [-1595, -1495, -995, 5], 5, "top"],
        ["up", [-1595, -1495, -995, 5], xr.DataArray(5, name="ssh"), "top"]
        ])
def test_grid_dz2depth(positive, expected, ref, ref_type):

    dz = xr.DataArray(
        np.resize([100, 500, 1000.], (2, 3)).T,
        dims=("z", "x"),
        coords={"z": ("z", np.arange(3, dtype="d"))}
        )

    depth = grid.dz2depth(dz, positive, ref=ref, ref_type=ref_type)
    np.testing.assert_allclose(depth.isel(x=1), expected)
    assert depth.z[0] == -0.5

    depth = grid.dz2depth(dz, positive, ref=ref, ref_type=ref_type, centered=True)
    assert depth[0, 0] == 0.5 * sum(expected[:2])
    assert depth.z[0] == 0


def test_coords_decode_cf_dz2depth():

    ds = xoa.open_data_sample("hycom.gdp.h.nc")
    ds = ds.rename(h="dz")
    dsd = grid.decode_cf_dz2depth(ds)
    assert "depth" in dsd.coords


def test_grid_torect():

    x = xr.DataArray(np.arange(4), dims='x')
    y = xr.DataArray(np.arange(3), dims='y')
    lon = xr.DataArray(np.ones((3, 4)), dims=('y', 'x'), coords={"y": y, "x": x})
    lat = xr.DataArray(np.ones((3, 4)), dims=('y', 'x'), coords={"y": y, "x": x})
    temp = xr.DataArray(
        np.ones((2, 3, 4)), dims=('time', 'y', 'x'),
        coords={'lon': lon, 'lat': lat, "y": y, "x": x})

    tempr = grid.to_rect(temp)
    assert tempr.dims == ('time', 'lat', 'lon')
    assert tempr.lon.ndim == 1
    np.testing.assert_allclose(tempr.lon.values, temp.lon[0].values)

    ds = xr.Dataset({"temp": temp})
    dsr = grid.to_rect(ds)
    assert dsr.temp.dims == tempr.dims



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# %% Path setup

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.join(os.path.abspath(
    os.path.dirname(__file__)), 'ext'))


# %% Project information

project = 'xoa'
copyright = '2020-2021, Shom'
author = 'Shom'

# The full version, including alpha/beta/rc tags
import xoa
release = xoa.__version__
xoa.register_accessors(xoa=True, xcf=True, decode_sigma=True)


# %% General configuration

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    "sphinx.ext.extlinks",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    'genoptions',
    'gencfspecs',
    'sphinxarg.ext',
    'sphinxcontrib.programoutput',
    'sphinx_autosummary_accessors',
    'sphinx_gallery.gen_gallery',
    'xoa.cfgm'
]

# Add any paths that contain templates here, relative to this directory.
import sphinx_autosummary_accessors
templates_path = ['_templates', sphinx_autosummary_accessors.templates_path]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['xoa.']


# %% Options for HTML output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# %% Autosumarry
autosummary_generate = True

# %% Intersphinx
intersphinx_mapping = {
    'python': ('https://docs.python.org/fr/3/', None),
    'cmocean': ('https://matplotlib.org/cmocean/', None),
    'configobj': ('https://configobj.readthedocs.io/en/latest/', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'numba:': ('https://numba.readthedocs.io/en/stable/', None),
    'numpy': ("https://numpy.org/doc/stable/", None),
    'pandas': ('https://pandas.pydata.org/docs/', None),
    'proplot': ('https://proplot.readthedocs.io/en/latest/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'xarray': ('http://xarray.pydata.org/en/stable/', None),
    # 'xesmf': ("https://xesmf.readthedocs.io/en/latest/", None)
    }

# %% Napoleon
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = False
napoleon_use_rtype = False

# %% Genoptions
genoptions_table = 'genoptions/table.txt'
genoptions_declarations = 'genoptions/declarations.txt'

# %% Cfgm
import xoa.cf
cfgm_get_cfgm_func = xoa.cf._get_cfgm_
cfgm_rst_file = "cf.txt"

# %% Extlinks
extlinks = {
    "issue": ("https://github.com/pydata/xarray/issues/%s", "GH"),
    "pull": ("https://github.com/pydata/xarray/pull/%s", "PR"),
}

# %% Nbsphinx
#nbsphinx_timeout = 120  # in seconds

# %% Sphinx gallery
sphinx_gallery_conf = {
    "examples_dirs": "../examples",
    "gallery_dirs": "examples",
    "binder": {
        'org': 'VACUMM',
        'repo': 'xoa',
        'branch': 'master',
        'binderhub_url': 'https://mybinder.org',
        'dependencies': [
            './binder/environment.yml',
            './binder/apt.txt',
            './binder/setup.py'
            ],
        'notebooks_dir': 'notebooks',
        'use_jupyter_lab': True,
        },
    }

# %% User directives

def setup(app):

    app.add_css_file('custom.css')

    app.add_object_type('confopt', 'confopt',
                        objname='configuration option',
                        indextemplate='pair: %s; configuration option')
    app.add_object_type('confsec', 'confsec',
                        objname='configuration section',
                        indextemplate='pair: %s; configuration section')
    app.add_object_type('confval', 'confval',
                        objname='sphinx configuration value',
                        indextemplate='pair: %s; sphinx configuration value')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
xarray-based ocean analysis library

The successor of Vacumm.
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import warnings
import platform

import pkg_resources
import appdirs
import configobj
import validate


# Taken from xarray
try:
    __version__ = pkg_resources.get_distribution("xoa").version
except Exception:
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"

_RE_OPTION_MATCH = re.compile(r"^(\w+)\W(\w+)$").match

#: Specifications of configuration options
CONFIG_SPECS = """
[cf] # cf module
cache=boolean(default=True) # use the :mod:`~xoa.cf` in memory and file caches

[plot] # plot parameters
cmapdiv = string(default="cmo.balance") # defaut diverging colormap
cmappos = string(default="cmo.amp")     # default positive colormap
cmapneg = string(default="cmo.tempo_r") # default negative colormap
cmapcyc = string(default="cmo.phase")   # default cyclic colormap

"""

#: Default xoa user configuration file
DEFAULT_USER_CONFIG_FILE = os.path.join(
    appdirs.user_config_dir("xoa"), "xoa.cfg"
)

# Directory of sample files
_SAMPLE_DIR = os.path.join(os.path.dirname(__file__), '_samples')

_PACKAGES = [
    "appdirs",
    "cartopy",
    "cmocean",
    "configobj",
    "matplotlib",
    "numpy",
    "pandas",
    "scipy",
    "xarray",
    "xesmf"
    ]


class XoaError(Exception):
    pass


class XoaConfigError(XoaError):
    pass


class XoaWarning(UserWarning):
    pass


def xoa_warn(message, stacklevel=2):
    """Issue a :class:`XoaWarning` warning

    Example
    -------
    .. ipython:: python
        :okwarning:

        @suppress
        from xoa import xoa_warn
        xoa_warn('Be careful!')
    """
    warnings.warn(message, XoaWarning, stacklevel=stacklevel)


def _get_cache_():
    from . import __init__
    if not hasattr(__init__, "_XOA_CACHE"):
        __init__._XOA_CACHE = {}
    return  __init__._XOA_CACHE


def load_options(cfgfile=None):
    """Load specified options

    Parameters
    ----------
    cfgfile: file, list(str), dict

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import load_options
        # Dict
        load_options({'plot': {'cmappos': 'mycmap'}})

        # Lines
        optlines = "[plot]\\n cmappos=mycmap".split('\\n')
        load_options(optlines)
    """
    _get_cache_()
    xoa_cache = _get_cache_()

    if "cfgspecs" not in xoa_cache:
        xoa_cache["cfgspecs"] = configobj.ConfigObj(
            CONFIG_SPECS.split("\n"),
            list_values=False,
            interpolation=False,
            raise_errors=True,
            file_error=True,
        )
    if "options" not in xoa_cache:
        xoa_cache["options"] = configobj.ConfigObj(
            (
                DEFAULT_USER_CONFIG_FILE
                if os.path.exists(DEFAULT_USER_CONFIG_FILE)
                else None
            ),
            configspec=xoa_cache["cfgspecs"],
            file_error=False,
            raise_errors=True,
            list_values=True,
        )
    if cfgfile:
        xoa_cache["options"].merge(
            configobj.ConfigObj(
                cfgfile, file_error=True, raise_errors=True, list_values=True
            )
        )
    xoa_cache["options"].validate(validate.Validator(), copy=True)


def _get_options_():
    xoa_cache = _get_cache_()
    if "options" not in xoa_cache:
        load_options()
    return xoa_cache["options"]


def get_option(section, option=None):
    """Get a config option

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import get_option
        print(get_option('plot', 'cmapdiv'))
        print(get_option('plot.cmapdiv'))
    """
    options = _get_options_()
    if option is None:
        m = _RE_OPTION_MATCH(section)
        if m:
            section, option = m.groups()
        else:
            raise XoaConfigError(
                "You must provide an option name to get_option"
            )
    try:
        value = options[section][option]
    except Exception:
        return XoaConfigError(f"Invalid section/option: {section}/{option}")
    return value


class set_options(object):
    """Set configuration options

    Parameters
    ----------
    section: str, None
    **options: dict
        If a key is in the format "<section>.<option>", then the section
        is overwritten.


    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import set_options, get_option

        # Classic: for the session
        set_options('plot', cmapdiv='cmo.balance', cmappos='cmo.amp')

        # With dict
        opts = {"plot.cmapdiv": "cmo.balance"}
        set_options(**opts)

        # Context: temporary
        with set_options('plot', cmapdiv='cmo.delta'):
            print('within context:', get_option('plot.cmapdiv'))
        print('after context:', get_option('plot.cmapdiv'))

    """

    def __init__(self, section=None, **options):
        # Format before being ingested
        self.xoa_cache = _get_cache_()
        self.old_options = self.xoa_cache.get("options")
        if "options" in self.xoa_cache:
            del self.xoa_cache["options"]
        opts = {}
        for option, value in options.items():
            m = _RE_OPTION_MATCH(option)
            if m:
                sec, option = m.groups()
            else:
                if section is None:
                    raise XoaConfigError(
                        "You must specify the section explicitly or through the option name")
                sec = section
            opts.setdefault(sec, {})[option] = value

        # Ingest options
        load_options(opts)

    def __enter__(self):
        return self.xoa_cache["options"]

    def __exit__(self, type, value, traceback):
        if self.old_options:
            self.xoa_cache["options"] = self.old_options
        else:
            del self.xoa_cache["options"]


def set_option(option, value):
    """Set a single option using the flat format, i.e ``section.option``

    Parameters
    ----------
    option: str
        Option name in the ``section.option`` format
    value:
        Value to set

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import set_option
        set_option('plot.cmapdiv', 'cmo.balance');
    """
    return set_options(None, **{option: value})


def reset_options():
    """Restore options to their default values in the current session

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import get_option, set_options, reset_options
        print(get_option('plot.cmapdiv'))
        set_options('plot', cmapdiv='mycmap')
        print(get_option('plot.cmapdiv'))
        reset_options()
        print(get_option('plot.cmapdiv'))
    """
    xoa_cache = _get_cache_()
    del xoa_cache['options']


def show_options(specs=False):
    """Print current xoa configuration

    Parameters
    ----------
    specs: bool
        Print option specifications instead

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import show_options
        show_options()
        show_options(specs=True)
    """
    if specs:
        print(CONFIG_SPECS.strip("\n"))
    else:
        print("\n".join(_get_options_().write())
              .strip("\n").replace('#', ' #'))


def _parse_requirements_(reqfile):
    re_match_specs_match = re.compile(r"^(\w+)(\W+.+)?$").match
    reqs = {}
    with open(reqfile) as f:
        for line in f:
            line = line.strip().strip("\n")
            if line and not line.startswith("#"):
                m = re_match_specs_match(line)
                if m:
                    reqs[m.group(1)] = m.group(2)
    return reqs


def show_versions():
    """Print the versions of xoa and of some dependencies

    Example
    -------
    .. ipython:: python
        :okexcept:

        @suppress
        from xoa import show_versions
        show_versions()
    """
    print("- python:", platform.python_version())
    print("- xoa:", __version__)
    for package in _PACKAGES:
        try:
            version = pkg_resources.get_distribution(package).version
        except pkg_resources.DistributionNotFound:
            version = "NOT INSTALLED or UKNOWN"
        print(f"- {package}: {version}")


def show_paths():
    """Print some xoa paths

    Example
    -------
    .. ipython:: python
        :okexcept:

        @suppress
        from xoa import show_paths
        show_paths()
    """
    print("- xoa library dir:", os.path.dirname(__file__))
    from . import cf
    asterix = False
    for label, path in [("user config file", DEFAULT_USER_CONFIG_FILE),
                        ("user CF specs file", cf.USER_CF_FILE),
                        ("user CF cache file", cf.USER_CF_CACHE_FILE)]:
        if not os.path.exists(path):
            asterix = True
            path = path + " [*]"
        print("-", label+":", path)
    print("- data samples:", " ".join(get_data_sample()))
    if asterix:
        print("*: file not present")


def show_info(opt_specs=True):
    """Print xoa related info

    Example
    -------
    .. ipython:: python
        :okexcept:

        @suppress
        from xoa import show_info
        show_info()
    """
    print("# VERSIONS")
    show_versions()
    print("\n# FILES AND DIRECTORIES")
    show_paths()
    print("\n# OPTIONS")
    show_options(specs=opt_specs)


def get_data_sample(filename=None):
    """Get the absolute path to a sample file

    Parameters
    ----------
    filename: str, None
        Name of the sample. If ommited, a list of available samples
        name is returned.

    Returns
    -------
    str OR list(str)

    Example
    -------
    .. .ipython:: python

        @suppress
        from xoa import get_data_sample
        get_data_sample("croco.south-africa.surf.nc")
        get_data_sample()

    See also
    --------
    show_data_samples
    open_data_sample
    """
    if not os.path.exists(_SAMPLE_DIR):
        filenames = []
    else:
        filenames = os.listdir(_SAMPLE_DIR)
    if filename is None:
        return filenames
    if filename not in filenames:
        raise XoaError("Invalid data sample: "+filename)
    return os.path.join(_SAMPLE_DIR, filename)


def open_data_sample(filename, **kwargs):
    """Open a data sample with :func:`xarray.open_dataset` or :func:`pandas.read_csv`

    A shortcut to::

        xr.open_dataset(get_data_sample(filename))

    Parameters
    ----------
    filename: str
        File name of the sample

    Returns
    -------
    xarray.Dataset, pandas.DataFrame

    Example
    -------
    .. .ipython:: python

        @suppress
        from xoa import open_data_sample
        open_data_sample("croco.south-africa.nc")


    See also
    --------
    get_data_sample
    show_data_samples
    """
    fname = get_data_sample(filename)
    if fname.endswith("nc"):
        import xarray as xr
        return xr.open_dataset(fname, **kwargs)
    import pandas as pd
    return pd.read_csv(fname, **kwargs)


def show_data_samples():
    """Print the list of data samples

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa import show_data_samples
        show_data_samples()

    See also
    --------
    get_data_samples
    open_data_sample
    """
    print(' '.join(get_data_sample()))


def register_accessors(xoa=True, xcf=False, decode_sigma=False):
    """Register xarray accessors

    Parameters
    ----------
    xoa: bool, str
        Register the main accessors with
        :func:`~xoa.cf.register_xoa_accessors`.
    xcf: bool, str
        Register the :mod:`xoa.cf` module accessors with
        :func:`~xoa.cf.register_cf_accessors`.
    decode_sigma: bool, str
        Register the :mod:`xoa.sigma` module accessor with
        :func:`~xoa.cf.register_sigma_accessor`.

    See also
    --------
    xoa.accessors
    """
    if xoa:
        from .accessors import register_xoa_accessors
        kw = {"name": xoa} if isinstance(xoa, str) else {}
        register_xoa_accessors(**kw)
    if xcf:
        from .accessors import register_cf_accessors
        kw = {"name": xcf} if isinstance(xcf, str) else {}
        register_cf_accessors(**kw)
    if decode_sigma:
        from .accessors import register_sigma_accessor
        kw = {"name": decode_sigma} if isinstance(decode_sigma, str) else {}
        register_sigma_accessor(**kw)



# -*- coding: utf-8 -*-
"""
Test the :mod:`xoa.dyn` module
"""

import numpy as np
import xarray as xr

from xoa import dyn


def test_dyn_flow2d():

    lon = xr.DataArray(np.arange(4), dims='lon')
    lat = xr.DataArray(np.arange(3), dims='lat')
    u = xr.DataArray(lat*lon/3, dims=('lat', 'lon'), coords={'lon': lon, 'lat': lat})
    v = xr.DataArray((lat+lon)/2, dims=('lat', 'lon'), coords={'lon': lon, 'lat': lat})

    ff = dyn.flow2d(
        u, v, ([1., 2.], [1., 1.5]), np.timedelta64(3, "h"), np.timedelta64(2, "h"),
        date="2000-01-01")
    assert list(ff.coords) == ['lon', 'lat', 'time']
    assert ff.lon.dims == ('time', 'particles')
    assert ff.lon.shape == (3, 2)
    assert 'lat' in ff.coords
    np.testing.assert_allclose(
        ff.lon.values.ravel(),
        [1., 2., 1.0148674, 2.04497533, 1.02261437, 2.06865695])
    assert ff.time.values[0] == np.datetime64("2000-01-01")



"""Generate files to declare fortran functions"""

import os
import re
import logging
import importlib

from docutils.statemachine import string2lines
from sphinx.util.docutils import SphinxDirective


path_pat_mod_dir = os.path.join("{gendir}", "{mod_name}")
path_pat_mod_file = os.path.join(path_pat_mod_dir, "index.rst")
path_pat_func_file = os.path.join(path_pat_mod_dir, "{func_name}.rst")


def checkdir(path):
    pdir = os.path.dirname(path)
    if not os.path.exists(pdir):
        os.makedirs(pdir)


class GenFortran(SphinxDirective):

    has_content = True

    def run(self):

        if not self.content:
            return []

        # Loop on modules and descriptions
        rst_toctree = ".. toctree::\n\t:hidden:\n\n"
        rst_table = ".. list-table::\n\n"
        for mod_name_desc in self.content:
            smod = mod_name_desc.split(" ")
            mod_name = smod[0]
            mod_desc = " ".join(smod[1:])
            rst_toctree += f"\tgenfortran/{mod_name}/index\n"
            rst_table += f"\t* - :mod:`{mod_name}`\n"
            rst_table += f"\t  - {mod_desc}\n"

        # Insert toctree and tables
        rst_all = rst_toctree + "\n\n" + rst_table + "\n"
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        include_lines = string2lines(rst_all, convert_whitespace=1)
        self.state_machine.insert_input(include_lines, source)

        return []


re_directive_match = re.compile(
    r"^(?P<indent>\s*)\.\.\s+genfortran::\s*\n$").match
re_indent_match = re.compile(r"^(?P<indent>\s*)\S.+\n$").match


def generate_stub_files(srcdir, mod_name, mod_desc):

    gendir = os.path.join(srcdir, "genfortran")

    logging.info(f"Generating rst files for fortran wrapper "+mod_name)
    mod_content = importlib.import_module(mod_name)
    func_names = [func for func in dir(mod_content)
                  if not func.startswith('_')]

    # Write files
    mod_dir = path_pat_mod_dir.format(**locals())
    if not os.path.exists(mod_dir):
        os.makedirs(mod_dir)
    mod_file = path_pat_mod_file.format(**locals())
    with open(mod_file, "w") as f:
        f.write(mod_name + "\n" + len(mod_name)*"=" + "\n\n")
        f.write(mod_desc + "\n\n")
        f.write(f".. module:: {mod_name}\n\n")
        rst_table = ".. list-table::\n\n"
        rst_toctree = ".. toctree::\n\t:hidden:\n\n"
        for func_name in func_names:

            rst_table += f"\t* - :func:`{mod_name}.{func_name}`\n"
            func = getattr(mod_content, func_name)
            func_sig = func.__doc__.split("\n")[0]
            rst_table += f"\t  - {func_sig}\n"
            rst_toctree += f"\t{func_name}\n"

            with open(path_pat_func_file.format(
                    **locals()), "w") as ff:
                ff.write(func_name+"\n"+len(func_name)*"="+"\n\n")
                ff.write(f".. currentmodule:: {mod_name}\n\n")
                out, call = func_sig.split('=')
                ff.write(f".. autofunction:: {call}\n\n")
        f.write(rst_toctree+"\n\n")
        f.write(rst_table)


def parse_and_generate(app):
    """Parse rst files to find directives and generate stub files"""

    # Get file list
    env = app.builder.env
    srcdir = env.srcdir
    if app.config.genfortran_src_files:
        srcfiles = [os.path.join(srcdir, srcfile) for srcfile
                    in app.config.genfortran_src_files]
    else:
        env = app.builder.env
        srcfiles = [env.doc2path(x, base=None) for x in env.found_docs
                    if os.path.isfile(env.doc2path(x))]

    # Parse files
    for srcfile in srcfiles:
        if not os.path.exists(srcfile):
            logging.warning("[genfortran] file not found: "+srcfile)
            continue
        with open(srcfile) as f:
            indent = None
            for line in f:
                m = re_directive_match(line)
                if m:
                    indent = m.group('indent')
                    continue
                if indent is None:
                    continue
                m = re.match("^"+indent + r"\s+(?P<mod_name>[\w.]+)" +
                             r"(?P<mod_desc>\s.*)\n$", line)
                if m:
                    generate_stub_files(
                        srcdir, m.group("mod_name"),
                        m.group("mod_desc").strip())
                    continue
                m = re_indent_match(line)
                if m and len(m.group('indent')) <= len(indent):
                    indent = None


def setup(app):

    app.add_directive("genfortran", GenFortran)
    app.connect('builder-inited', parse_and_generate)
    app.add_config_value('genfortran_src_files', [], [], [list])

    return {'version': '0.1'}



"""
Low level interpolation routines accelerated with numba

The numerical inputs and outputs of all these routines are of scalar
or numpy.ndarray type.
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import math

import numpy as np
import numba

from .geo import _haversine_

NOT_CI = os.environ.get("CI", "false") == "false"


# %% 1D routines

@numba.njit(cache=NOT_CI)
def get_iminmax(data1d):
    """The first and last non nan values for a 1d array

    Parameters
    ----------
    data1d: array_like(n)

    Return
    ------
    int
        Index of the first valid value
    int
        Index of the last valid value
    """
    imin = -1
    imax = -1
    n = len(data1d)
    for i in range(n):
        if imin < 0 and not np.isnan(data1d[i]):
            imin = i
        if imax < 0 and not np.isnan(data1d[n-1-i]):
            imax = n-1-i
        if imax > 0 and imin > 0:
            break
    return imin, imax


@numba.njit(parallel=False, cache=NOT_CI)
def nearest1d(vari, yi, yo, extrap="no"):
    """Nearest interpolation of nD data along an axis with varying coordinates

    Warning
    -------
    `nxi` must be either a multiple or a divisor of `nxo`,
    and multiple of `nxiy`.

    Parameters
    ----------
    vari: array_like(nxi, nyi)
    yi: array_like(nxiy, nyi)
    yo: array_like(nxo, nyo)

    Return
    ------
    array_like(nx, nyo): varo
        With `nx=max(nxi, nxo)`
    """
    # Shapes
    nxi, nyi = vari.shape
    nxiy = yi.shape[0]
    nxi, nyi = vari.shape
    nxo, nyo = yo.shape
    nx = max(nxi, nxo)

    # Init output
    varo = np.full((nx, nyo), np.nan, dtype=vari.dtype)

    # Loop on the varying dimension
    for ix in numba.prange(nx):

        # Index along x for coordinate arrays
        ixi = min(nxi-1, ix % nxi)
        ixiy = min(nxiy-1, ix % nxiy)
        ixoy = min(nxo-1, ix % nxo)

        # Loop on input grid
        iyimin, iyimax = get_iminmax(yi[ixiy])
        iyomin, iyomax = get_iminmax(yo[ixoy])
        for iyi in range(iyimin, iyimax):

            # Out of bounds
            if yi[ixiy, iyi+1] < yo[ixoy, iyomin]:
                continue
            if yi[ixiy, iyi] > yo[ixoy, iyomax]:
                break

            # Loop on output grid
            for iyo in range(iyomin, iyomax+1):

                dy0 = yo[ixoy, iyo] - yi[ixiy, iyi]
                dy1 = yi[ixiy, iyi+1] - yo[ixoy, iyo]

                # Above
                if dy1 < 0:  # above
                    break

                # Below
                if dy0 < 0:
                    iyomin = iyo + 1

                # Interpolations
                elif dy0 <= dy1:
                    varo[ix, iyo] = vari[ixi, iyi]
                else:
                    varo[ix, iyo] = vari[ixi, iyi+1]

    # Extrapolation
    if extrap != "no":
        varo = extrap1d(varo, extrap)

    return varo


@numba.njit(parallel=False, cache=NOT_CI)
def linear1d(vari, yi, yo, extrap="no"):
    """Linear interpolation of nD data along an axis with varying coordinates

    Warning
    -------
    `nxi` must be either a multiple or a divisor of `nxo`,
    and multiple of `nxiy`.

    Parameters
    ----------
    vari: array_like(nxi, nyi)
    yi: array_like(nxiy, nyi)
    yo: array_like(nxo, nyo)

    Return
    ------
    array_like(nx, nyo): varo
        With `nx=max(nxi, nxo)`
    """
    # Shapes
    nxi, nyi = vari.shape
    nxiy = yi.shape[0]
    nxi, nyi = vari.shape
    nxo, nyo = yo.shape
    nx = max(nxi, nxo)

    # Init output
    varo = np.full((nx, nyo), np.nan, dtype=vari.dtype)

    # Loop on the varying dimension
    for ix in numba.prange(nx):

        # Index along x for coordinate arrays
        ixi = min(nxi-1, ix % nxi)
        ixiy = min(nxiy-1, ix % nxiy)
        ixoy = min(nxo-1, ix % nxo)

        # Loop on input grid
        iyimin, iyimax = get_iminmax(yi[ixiy])
        iyomin, iyomax = get_iminmax(yo[ixoy])
        for iyi in range(iyimin, iyimax):

            # Out of bounds
            if yi[ixiy, iyi+1] < yo[ixoy, iyomin]:
                continue
            if yi[ixiy, iyi] > yo[ixoy, iyomax]:
                break

            # Loop on output grid
            for iyo in range(iyomin, iyomax+1):

                dy0 = yo[ixoy, iyo] - yi[ixiy, iyi]
                dy1 = yi[ixiy, iyi+1] - yo[ixoy, iyo]

                # Above
                if dy1 < 0:  # above
                    break

                # Below
                if dy0 < 0:
                    iyomin = iyo + 1

                # Interpolation
                elif dy0 > 0 or dy1 > 0:

                    varo[ix, iyo] = (
                        (vari[ixi, iyi]*dy1 + vari[ixi, iyi+1]*dy0) /
                        (dy0+dy1))

    # Extrapolation
    if extrap != "no":
        varo = extrap1d(varo, extrap)

    return varo


@numba.njit(parallel=False, cache=NOT_CI)
def cubic1d(vari, yi, yo, extrap="no"):
    """Cubic interpolation of nD data along an axis with varying coordinates

    Warning
    -------
    `nxi` must be either a multiple or a divisor of `nxo`,
    and multiple of `nxiy`.

    Parameters
    ----------
    vari: array_like(nxi, nyi)
    yi: array_like(nxiy, nyi)
    yo: array_like(nxo, nyo)

    Return
    ------
    array_like(nx, nyo): varo
        With `nx=max(nxi, nxo)`
    """
    # Shapes
    nxi, nyi = vari.shape
    nxiy = yi.shape[0]
    nxi, nyi = vari.shape
    nxo, nyo = yo.shape
    nx = max(nxi, nxo)

    # Init output
    varo = np.full((nx, nyo), np.nan, dtype=vari.dtype)

    # Loop on the varying dimension
    for ix in numba.prange(nx):

        # Index along x for coordinate arrays
        ixi = min(nxi-1, ix % nxi)
        ixiy = min(nxiy-1, ix % nxiy)
        ixoy = min(nxo-1, ix % nxo)

        # Loop on input grid
        iyimin, iyimax = get_iminmax(yi[ixiy])
        iyomin, iyomax = get_iminmax(yo[ixoy])
        for iyi in range(iyimin, iyimax):

            # Out of bounds
            if yi[ixiy, iyi+1] < yo[ixoy, iyomin]:
                continue
            if yi[ixiy, iyi] > yo[ixoy, iyomax]:
                break

            # Loop on output grid
            for iyo in range(iyomin, nyo):

                dy0 = yo[ixoy, iyo] - yi[ixiy, iyi]
                dy1 = yi[ixiy, iyi+1] - yo[ixoy, iyo]

                # Above
                if dy1 < 0:  # above
                    break

                # Inside
                if dy0 >= 0 and dy1 >= 0:

                    iyomin = iyo
                    mu = dy0 / (dy0+dy1)

                    # Extrapolations
                    if iyi == iyimin:  # y0
                        vc0 = 2*vari[ix, iyi] - vari[ix, iyi+1]
                    else:
                        vc0 = vari[ixi, iyi-1]
                    if iyi == iyimax-1:  # y3
                        vc1 = 2*vari[ixi, iyi+1] - vari[ixi, iyi]
                    else:
                        vc1 = vari[ixi, iyi+2]

                    # Interpolation
                    varo[ix, iyo] = (vc1 - vari[ix, iyi+1]
                                     - vc0 + vari[ix, iyi])
                    varo[ix, iyo] = (
                        mu**3*varo[ix, iyo] +
                        mu**2*(vc0 - vari[ix, iyi] - varo[ix, iyo]))
                    varo[ix, iyo] += mu*(vari[ix, iyi+1] - vc0)
                    varo[ix, iyo] += vari[ix, iyi]

    # Extrapolation
    if extrap != "no":
        varo = extrap1d(varo, extrap)

    return varo


@numba.njit(parallel=False, cache=NOT_CI)
def hermit1d(vari, yi, yo, extrap="no", bias=0., tension=0.):
    """Hermitian interp. of nD data along an axis with varying coordinates

    Warning
    -------
    `nxi` must be either a multiple or a divisor of `nxo`,
    and multiple of `nxiy`.

    Parameters
    ----------
    vari: array_like(nxi, nyi)
    yi: array_like(nxiy, nyi)
    yo: array_like(nxo, nyo)
    bias: float
    tension: float

    Return
    ------
    array_like(nx, nyo): varo
        With `nx=max(nxi, nxo)`
    """
    # Shapes
    nxi, nyi = vari.shape
    nxiy = yi.shape[0]
    nxi, nyi = vari.shape
    nxo, nyo = yo.shape
    nx = max(nxi, nxo)

    # Init output
    varo = np.full((nx, nyo), np.nan, dtype=vari.dtype)

    # Loop on the varying dimension
    for ix in numba.prange(nx):

        # Index along x for coordinate arrays
        ixi = min(nxi-1, ix % nxi)
        ixiy = min(nxiy-1, ix % nxiy)
        ixoy = min(nxo-1, ix % nxo)

        # Loop on input grid
        iyimin, iyimax = get_iminmax(yi[ixiy])
        iyomin, iyomax = get_iminmax(yo[ixoy])
        for iyi in range(iyimin, iyimax):

            # Out of bounds
            if yi[ixiy, iyi+1] < yo[ixoy, iyomin]:
                continue
            if yi[ixiy, iyi] > yo[ixoy, iyomax]:
                break

            # Loop on output grid
            for iyo in range(iyomin, nyo):

                dy0 = yo[ixoy, iyo] - yi[ixiy, iyi]
                dy1 = yi[ixiy, iyi+1] - yo[ixoy, iyo]

                # Above
                if dy1 < 0:  # above
                    break

                # Inside
                if dy0 >= 0 and dy1 >= 0:

                    iyomin = iyo
                    mu = dy0 / (dy0+dy1)

                    # Extrapolations
                    if iyi == iyimin:  # y0
                        vc0 = 2*vari[ix, iyi] - vari[ix, iyi+1]
                    else:
                        vc0 = vari[ixi, iyi-1]
                    if iyi == iyimax-1:  # y3
                        vc1 = 2*vari[ixi, iyi+1] - vari[ixi, iyi]
                    else:
                        vc1 = vari[ixi, iyi+2]

                    # Interpolation
                    mu = dy0 / (dy0+dy1)
                    a0 = 2*mu**3 - 3*mu**2 + 1
                    a1 = mu**3 - 2*mu**2 + mu
                    a2 = mu**3 - mu**2
                    a3 = -2*mu**3 + 3*mu**2
                    varo[ix, iyo] = a0*vari[ix, iyi]
                    varo[ix, iyo] += a1*(
                        (vari[ix, iyi]-vc0) *
                        (1+bias)*(1-tension)/2 +
                        (vari[ix, iyi+1]-vari[ix, iyi]) *
                        (1-bias)*(1-tension)/2)
                    varo[ix, iyo] += a2*(
                        (vari[ix, iyi+1]-vari[ix, iyi]) *
                        (1+bias)*(1-tension)/2 +
                        (vc1-vari[ix, iyi+1]) *
                        (1-bias)*(1-tension)/2)
                    varo[ix, iyo] += a3*vari[ix, iyi+1]

    if extrap != "no":
        varo = extrap1d(varo, extrap)

    return varo


@numba.njit(parallel=False)
def extrap1d(vari, mode):
    """Extrapolate valid data to the top and/or bottom

    Parameters
    ----------
    vari: array_like(nx, ny)
    mode: {"top", "bottom", "both", "no"}
        Extrapolation mode

    Return
    ------
    array_like(nx, ny): varo
    """
    varo = vari.copy()
    if mode == "no":
        return varo
    nx, ny = vari.shape

    # Loop on varying dim
    for ix in numba.prange(0, nx):
        iybot, iytop = get_iminmax(vari[ix])
        if iybot == -1:
            continue
        if mode == "both" or mode == "bottom":
            varo[ix, :iybot] = varo[ix, iybot]
        if mode == "both" or mode == "top":
            varo[ix, iytop+1:] = varo[ix, iytop]

    return varo


@numba.njit(parallel=False, cache=NOT_CI)
def cellave1d(vari, yib, yob, extrap="no", conserv=False):
    """Cell average regrid. of nD data along an axis with varying coordinates

    Warning
    -------
    `nxi` must be either a multiple or a divisor of `nxo`,
    and multiple of `nxiy`.

    Parameters
    ----------
    vari: array_like(nxi, nyi)
    yib: array_like(nxiy, nyi+1)
    yob: array_like(nxo, nyo+1)

    Return
    ------
    array_like(nx, nyo): varo
        With `nx=max(nxi, nxo)`
    """
    # Shapes
    nxi, nyib = vari.shape
    nxiy, nyi = yib.shape
    nxi, nyi = vari.shape
    nxo, nyob = yob.shape
    nx = max(nxi, nxo)
    nyo = nyob - 1

    # Init output
    varo = np.zeros((nx, nyo), dtype=vari.dtype)

    # Loop on the varying dimension
    for ix in numba.prange(nx):

        # Index along x for coordinate arrays
        ixi = min(nxi-1, ix % nxi)
        ixiy = min(nxiy-1, ix % nxiy)
        ixoy = min(nxo-1, ix % nxo)

        # Loop on output cells to be filled
        iyi0 = 0
        for iyo in range(nyo):

            if yob[ixoy, iyo] == yob[ixoy, iyo+1]:
                continue

            # Loop on input cells
            wo = 0.
            for iyi in range(iyi0, nyi):

                # Current input bounds
                yib0 = yib[ixiy, iyi]
                yib1 = yib[ixiy, iyi+1]

                # Extrapolation
                if((extrap == "bellow" or extrap == "both") and
                   iyi == 0 and yib0 > yob[ixoy, iyo]):
                    yib0 = yob[ixoy, iyo]
                if((extrap == "above" or extrap == "both") and
                   iyi == nyi-1 and yib1 < yob[ixoy, iyo+1]):
                    yib1 = yob[ixoy, iyo+1]

                # No intersection
                if yib0 > yob[ixoy, iyo+1]:
                    break
                if yib1 < yob[ixoy, iyo]:
                    iyi0 = iyi + 1
                    continue

                # Contribution of intersection
                dyio = min(yib1, yob[ixoy, iyo+1]) - max(yib0, yob[ixoy, iyo])
                if conserv and yib0 != yib1:
                    dyio = dyio / (yob[ixoy, iyo+1] - yob[ixoy, iyo])
                if not np.isnan(vari[ixi, iyi]):
                    wo = wo + dyio
                    varo[ixi, iyo] += vari[ix, iyi] * dyio

                # Next input cell?
                if yib1 >= yob[ixoy, iyo+1]:
                    break

            # Normalize
            if not conserv:
                if wo != 0:
                    varo[ix, iyo] /= wo
                else:
                    varo[ix, iyo] = np.nan

    return varo


# %% 2D routines

@numba.njit(fastmath=True)
def closest2d(xxi, yyi, xo, yo):
    """Find indices of closest point on 2D lon/lat grid

    Parameters
    ----------
    xxi: array_like(nyi, nxi)
        Grid longitudes in degrees
    yyi: array_like(nyi, nxi)
        Grid latitudes in degrees
    xo:
        Point longtitude
    yo:
        Point latitude

    Return
    ------
    int: i
        index along second dim
    int: j
        Index along first dim
    """
    nyi, nxi = xxi.shape
    mindist = np.pi
    i = 0
    j = 0
    for jt in range(0, nyi):
        for it in range(0, nxi):
            dist = _haversine_(xo, yo, xxi[jt, it], yyi[jt, it])
            if dist <= mindist:
                i = it
                j = jt
                mindist = dist
    return i, j


@numba.njit(fastmath=True)
def cell2relloc(x1, x2, x3, x4, y1, y2, y3, y4, x, y):
    """Compute coordinates of point relative to a curvilinear cell

    Cell shape::

      2 - 3
      |   |
      1 - 4

    Example
    -------
    >>> cell2relloc(0., -2., 0., 2., 0., 1., 1., 0., 0., 0.5)
    (0.5, 0.5)


    See also
    --------
    http://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/19890018062_1989018062.pdf
    """
    small = np.finfo(np.float64).eps * 2

    # Coefs
    a = x4 - x1
    b = x2 - x1
    c = x3 - x4 - x2 + x1
    d = y4 - y1
    e = y2 - y1
    f = y3 - y4 - y2 + y1

    # Solve A*p**2 + B*p + C = 0
    yy = y - y1
    xx = x - x1
    AA = c*d - a*f
    BB = -c*yy + b*d + xx*f - a*e
    CC = -yy*b + e*xx
    if abs(AA) < small:
        p1 = -CC / BB
        p2 = p1
    else:
        DD = BB**2 - 4*AA*CC
        sDD = math.sqrt(DD)
        p1 = (-BB-sDD) / (2*AA)
        p2 = (-BB+sDD) / (2*AA)

    # Get q from p
    if abs(b+c*p1) > small:
        q1 = (xx-a*p1)/(b+c*p1)
    else:
        q1 = (yy-d*p1)/(e+f*p1)

    # Select point closest to center
    if p1 < 0. or p1 > 1. or q1 < 0. or q1 > 1.:
        if abs(b+c*p2) > small:
            q2 = (xx-a*p2) / (b+c*p2)
        else:
            q2 = (yy-d*p2) / (e+f*p2)
        p = p2
        q = q2
    else:
        p = p1
        q = q1

    if p < -small or q < -small:
        p = -1.
        q = -1.

    return p, q


@numba.njit(fastmath=True, cache=NOT_CI)
def grid2relloc(xxi, yyi, xo, yo):
    """Compute coordinates of point relative to a curvilinear grid

    Parameters
    ----------
    xxi: array_like(nyi, nxi)
        Grid longitudes in degrees
    yyi: array_like(nyi, nxi)
        Grid latitudes in degrees
    xo: float
        Point longitude
    yo: float
        Point latitude

    Return
    ------
    float:
        The integer part gives the grid cell index along the second dim,
        and the fractional part gives the coordinate relative this cell.
        A value of -1 means outside the grid.
    float:
        The integer part gives the grid cell index along the first dim,
        and the fractional part gives the coordinate relative this cell.
        A value of -1 means outside the grid.
    """

    p = -1.
    q = -1.
    small = np.finfo(np.float64).eps
    nyi, nxi = xxi.shape

    # Find the closest corner
    ic, jc = closest2d(xxi, yyi, xo, yo)

    # Curvilinear to rectangular with a loop on four candidate cells
    for j in range(max(jc-1, 0), min(jc+1, nyi-1)):
        for i in range(max(ic-1, 0), min(ic+1, nxi-1)):

            # Get relative position
            a, b = cell2relloc(
                xxi[j, i], xxi[j+1, i], xxi[j+1, i+1], xxi[j, i+1],
                yyi[j, i], yyi[j+1, i], yyi[j+1, i+1], yyi[j, i+1],
                xo, yo)

            # Store absolute indices
            if ((a >= 0.-small) and (a <= 1.+small) and
                    (b >= 0.-small) and (b <= 1.+small)):
                p = np.float64(i) + a
                q = np.float64(j) + b
                return p, q
    return p, q


@numba.njit(fastmath=True, cache=NOT_CI)
def grid2rellocs(xxi, yyi, xo, yo):
    """Compute coordinates of points relative to a curvilinear grid

    Parameters
    ----------
    xxi: array_like(nyi, nxi)
        Grid longitudes in degrees
    yyi: array_like(nyi, nxi)
        Grid latitudes in degrees
    xo: array_like(no)
        Point longitude
    yo: array_like(no)
        Point latitude

    Return
    ------
    array_like(no):
        The integer part gives the grid cell index along the second dim,
        and the fractional part gives the coordinate relative this cell.
        A value of -1 means outside the grid.
    array_like(no):
        The integer part gives the grid cell index along the first dim,
        and the fractional part gives the coordinate relative this cell.
        A value of -1 means outside the grid.
    """
    no = xo.size
    pp = np.zeros(no)
    qq = np.zeros(no)
    for i in range(xo.size):
        pp[i], qq[i] = grid2relloc(xxi, yyi, xo[i], yo[i])
    return pp, qq


@numba.njit(parallel=False, cache=NOT_CI)
def grid2locs(xxi, yyi, zzi, ti, vi, xo, yo, zo, to):
    """Linear interpolation of gridded data to random positions

    Parameters
    ----------
    xxi: array_like(nyi, nxi)
        Input grid longitudes in degrees, with `nyi==1` for 1D coordinates.
    yyi: array_like(nyi, nxi)
        Input grid latitudes in degrees, with `nxi==1` for 1D coordinates.
    zzi: array_like(nexz, nti, nzi, nyi, nxi)
        Input grid depths, positive up. Non effective dimensions
        must be set to 1. `nexz` may be equal or a multiple of `nex`.
    ti:  array_like(nti)
        Input times
    vi: array_like(nexz, ntiz, nzi, nyiz, nxiz)
        Input values.
    xo: array_like(no)
        Points longitude
    yo: array_like(no)
        Points latitude
    zo: array_like(no)
        Points depth, positive up.
    to: array_like(no)
        Points time.

    Return
    ------
    array_like(nex, no)
        Points value.
    """
    # Dimensions
    nyix, nxi = xxi.shape
    nyi, nxiy = yyi.shape
    nexz, ntiz, nzi, nyiz, nxiz = zzi.shape
    nexv, nti, nzi, nyi, nxi = vi.shape
    no = xo.shape[0]

    # Initalisations
    nex = max(nexv, nexz)
    vo = np.full((nex, no), np.nan, dtype=vi.dtype)
    bmask = np.isnan(vi)
    masko = np.isnan(xo) | np.isnan(yo) | np.isnan(zo) | np.isnan(to)
    ximin = xxi.min()
    ximax = xxi.max()
    yimin = yyi.min()
    yimax = yyi.max()
    zimin = zzi.min()
    zimax = zzi.max()
    timin = ti.min()
    timax = ti.max()
    curved = nyix != 1

    # Verifications
    assert not curved or (nxi == nxiy and nyi == nyix), (
        "linear4dto1: Invalid curved dimensions")
    assert nxiz == 1 or nxiz == nxi, "grid2locs: Invalid nxiz dimension"
    assert nyiz == 1 or nyiz == nyi, "grid2locs: Invalid nyiz dimension"
    assert ntiz == 1 or ntiz == nti, "grid2locs: Invalid ntiz dimension"

    # Loop on ouput points
    for io in numba.prange(no):
    # for io in range(no):

        if masko[io]:
            continue

        if ((nxi != 1 and (xo[io] < ximin or xo[io] > ximax)) or
                (nyi != 1 and (yo[io] < yimin or yo[io] > yimax)) or
                (nzi != 1 and (zo[io] < zimin or zo[io] > zimax)) or
                (nti != 1 and (to[io] < timin or to[io] > timax))):
            continue

        # Weights
        if curved:

            p, q = grid2relloc(xxi, yyi, xo[io], yo[io])
            if p < 0 or p > nxi-1 or q < 0 or q > nyi-1:
                continue  # outside the grid
            i = int(p)
            j = int(q)
            a = p - i
            b = q - j
            npi = 2
            npj = 2

        else:
            # - X
            if nxi == 1:
                i = 0
                a = 0.
                npi = 1
            elif xxi[0, nxi-1] == xo[io]:
                i = nxi - 1
                a = 0.
                npi = 1
            else:
                # i = minloc(xxi[1, :], dim=1, mask=xxi(1,:)>xo[io])-1
                i = np.searchsorted(xxi[0, :], xo[io], "right") - 1
                a = xo[io] - xxi[0, i]
                if abs(a) > 180.:
                    a -= 180.  # FIXME: grid2locs: abs(a)>180.
                a = a / (xxi[0, i+1] - xxi[0, i])
                npi = 2

            # - Y
            if nyi == 1:
                j = 0
                b = 0.
                npj = 1
            elif yyi[nyi-1, 0] == yo[io]:
                j = nyi - 1
                b = 0.
                npj = 1
            else:
                # j = minloc(yyi[:,1], dim=1, mask=yyi[:,1]>yo[io]) - 1
                j = np.searchsorted(yyi[:, 0], yo[io], "right") - 1
                b = (yo[io]-yyi[j, 0]) / (yyi[j+1, 0]-yyi[j, 0])
                npj = 2

        # - T
        if nti == 1:
            l = 0
            d = 0.
            npl = 1
        elif ti[nti-1] == to[io]:
            l = nti - 1
            d = 0.
            npl = 1
        else:
            l = np.searchsorted(ti, to[io], "right") - 1
            # l = minloc(ti, dim=1, mask=ti>to[io])-1
            if ti[l+1] == ti[l]:
                d = 0.
                npl = 1
            else:
                d = (to[io]-ti[l]) / (ti[l+1]-ti[l])
                npl = 2

        # - Z
        c = np.zeros(nexz)
        k = np.zeros(nexz, 'l')
        npk = np.zeros(nexz, 'l')
        if nzi == 1:
            k[:] = 0
            c[:] = 0.
            npk[:] = 1
        else:

            # Local zi

            if nxiz == 1:
                npiz = 1
                az = 0.
                iz = 0
            else:
                npiz = npi
                az = a
                iz = i

            if nyiz == 1:
                npjz = 1
                bz = 0.
                jz = 0
            else:
                npjz = npj
                bz = b
                jz = j

            if ntiz == 1:
                nplz = 1
                dz = 0.
                lz = 0
            else:
                nplz = npl
                dz = d
                lz = l

            zi = np.zeros((nexz, nzi))
            for ie in range(nexz):
                for ll in range(nplz):
                    for kk in range(nzi):
                        for jj in range(0, npjz):
                            for ii in range(0, npiz):
                                zi[ie, kk] += (
                                    zzi[ie, lz+ll, kk, jz+jj, iz+ii] *
                                    ((1-az) * (1-ii) + az * ii) *
                                    ((1-bz) * (1-jj) + bz * jj) *
                                    ((1-dz) * (1-ll) + dz * ll))

            # Normal stuff (c(nexz),zi[nexz,nzi),k(nexz)
            for ie in range(nexz):  # extra dim
                if zi[ie, nzi-1] == zo[io]:
                    k[ie] = nzi - 1
                    c[ie] = 0.
                    npk[ie] = 1
                else:
                    k[ie] = np.searchsorted(zi[ie], zo[io], "right") - 1
                    if zi[ie, k[ie]+1] == zi[ie, k[ie]]:
                        c[ie] = 0.
                        npk[ie] = 1
                    else:
                        c[ie] = ((zo[io]-zi[ie, k[ie]]) /
                                  (zi[ie, k[ie]+1] - zi[ie, k[ie]]))
                        npk[ie] = 2

        # Interpolate
        for ie in range(nex):
            if not bmask[
                    ie % nexv, l:l+npl,
                    k[ie % nexz]:k[ie % nexz]+npk[ie % nexz],
                    j:j+npj, i:i+npi].any():
                vo[ie % nex, io] = 0.
                for ll in range(npl):
                    for kk in range(npk[ie % nexz]):
                        for jj in range(npj):
                            for ii in range(npi):
                                vo[ie % nex, io] = (
                                    vo[ie % nex, io] +
                                    vi[ie % nex,
                                        l+ll, k[ie % nexz]+kk, j+jj, i+ii] *
                                    ((1-a) * (1-ii) + a * ii) *
                                    ((1-b) * (1-jj) + b * jj) *
                                    ((1-c[ie % nexz]) *
                                      (1-kk) + c[ie % nexz] * kk) *
                                    ((1-d) * (1-ll) + d * ll))

    return vo



"""
Filtering utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import xarray as xr

from .__init__ import XoaError, xoa_warn
from . import coords as xcoords


def get_window_func(window, *args, **kwargs):
    """Get a window function from its name

    Parameters
    ----------
    window: str, callable, list, array_like
        Specification to get the window function:

        - `callable`: used as is.
        - `str`: supposed to be function name of the :mod:`numpy` or
          :mod:`scipy.signal.windows` modules.
        - `list`, `array_like`: transformed to an array and interpolated
          onto ``size`` points.

    args, kwargs:
        Argument passed to the low level window function at calling time.

    Return
    ------
    callable
        A function that only a ``size`` argument

    Example
    -------
    .. ipython:: python

        @suppress
        import matplotlib.pyplot as plt
        @suppress
        from xoa.filter import get_window_func
        func0 = get_window_func("gaussian", 22, sym=True)
        func1 = get_window_func([1, 2, 5, 2, 1])
        plt.plot(func0(100), label='Gaussian');
        plt.plot(func1(100), label='List/array');
        @savefig api.filter.get_window_func.png
        plt.legend();
    """
    # Explicit values so create a wrapper that interpolate them
    if isinstance(window, (list, np.ndarray)):
        def window_func(size):
            x = np.linspace(0, 1, size)
            xp = np.linspace(0, 1, len(window))
            return np.interp(x, xp, window)
        return window_func

    # Base function
    if isinstance(window, str):
        if hasattr(np, window):
            func = getattr(np, window)
        else:
            import scipy.signal.windows as sw
            func = getattr(sw, window, None)
            if func is None:
                raise XoaError(f'Invalid window name: {window}')
    else:
        func = window

    # Wrapper with args
    if args or kwargs:
        def window_func(size):
            return func(size, *args, **kwargs)
        return window_func
    return func


# def get_window_kernel(size, window, *args, **kwargs):
#     """Generate an 1d kernel from a window name or function

#     It first get the window function, then returns::

#         window_func(size)

#     Parameters
#     ----------
#     size: int
#         Size of the output kernel
#     window: str, callable, list, array_like
#         Specification to get the window function:

#         - `callable`: used as is.
#         - `str`: supposed to be function name of the :mod:`numpy` or
#           :mod:`scipy.signal.windows` modules.
#         - `list`, `array_like`: transformed to an array and interpolated
#           onto ``size`` points.

#     Return
#     ------
#     np.ndarray
#         1D kernel

#     See also
#     --------
#     get_window_func
#     """
#     if isinstance(window, str):
#         window = get_window_func(window, *args, **kwargs)
#     elif isinstance(window, (list, np.ndarray)):
#         x = np.linspace(0, 1, size)
#         xp = np.linspace(0, 1, len(window))
#         return np.interp(x, xp, window)
#     return window(size)


def generate_isotropic_kernel(shape, window_func, fill_value=0, npt=None):
    """Generate an nD istropic kernel given a shape and a window function

    Parameters
    ----------
    shape: int, tuple
        Shape of the desired kernel
    window_func: str, callable
        Function that take a size parameter as unique argument.
        If a string, it is expected to be a numpy window function.
    fill_value: float
        Value to set when outside window bounds, near the corners
    npt: int, None
        Number of interpolation point to get the window value
        at all positions. It is infered from shape if not given.

    Return
    ------
    np.ndarray
        Isotropic kernel array

    Example
    -------

    .. ipython:: python

        @suppress
        import numpy as np, matplotlib.pyplot as plt
        @suppress
        from xoa.filter import generate_isotropic_kernel
        kernel = generate_isotropic_kernel((20, 30), "bartlett", np.nan)
        plt.matshow(kernel, cmap="cmo.solar");
        @savefig api.filter.generate_isotropic_kernel.png
        plt.colorbar();

    See also
    --------
    generate_orthogonal_kernel
    generate_kernel
    numpy.bartlett
    numpy.blackman
    numpy.hamming
    numpy.hanning
    numpy.kaiser
    numpy.interp

    """
    # Window function
    window_func = get_window_func(window_func)

    # Normalised indices
    indices = np.indices(shape).astype('d')
    for i, width in enumerate(shape):
        indices[i] /= (width-1)
        indices[i] -= 0.5

    # Distance from bounds with 0.5 at center and < 0 outside bounds
    x = 0.5 - np.sqrt((indices**2).sum(axis=0))

    # Window values
    if npt is None:
        npt = 2*max(shape)
    fp = window_func(npt)
    xp = np.linspace(0, 1, npt)

    # Interpolation
    kernel = np.interp(x.ravel(), xp, fp).reshape(x.shape)
    kernel[x < 0] = fill_value

    return kernel


def generate_orthogonal_kernel(kernels, window_func="ones", fill_value=0.):
    """Generate an nD kernel from orthogonal 1d kernels

    Parameters
    ----------
    kernels: list
        List of scalars and/or 1d kernels.
        In case of a scalar, it is converted to an 1d kernel with
        the ``window_func`` parameter.
    window_func: str, callable
        Function that take a size parameter as unique argument.
        If a string, it is expected to be a numpy window function.
    fill_value: float
        Value to set when outside window bounds, when creating an 1d kernel
        from a floating size

    Return
    ------
    np.ndarray
        Orthogonal kernel array

    Example
    -------
    .. ipython:: python

        @suppress
        import numpy as np, matplotlib.pyplot as plt
        @suppress
        from matplotlib.gridspec import GridSpec
        @suppress
        from xoa.filter import generate_orthogonal_kernel
        # From sizes and functions
        ny, nx = (21, 31)
        kernel = generate_orthogonal_kernel((ny, nx), "bartlett")
        j, i = 5, 20
        plt.plot([3, 4])
        fig = plt.figure(constrained_layout=True)
        gs = GridSpec(3, 3, figure=fig)
        ax0 = fig.add_subplot(gs[1:, :2])
        ax0.matshow(kernel)
        ax0.axhline(j, color='tab:red')
        ax1 = fig.add_subplot(gs[0, :2], sharex=ax0)
        ax1.plot(np.arange(nx), kernel[j], color='tab:red')
        ax2 = fig.add_subplot(gs[1:, -1], sharey=ax0)
        ax2.plot(kernel[:, i], np.arange(ny), color='tab:orange');
        @savefig api.filter.generate_orthogonal_kernel_0.png
        ax0.axvline(i, color='tab:orange')
        # From 1d kernels
        kernel = generate_orthogonal_kernel(([1, 1, 1], [1, 2, 3, 2, 1]))
        plt.figure()
        @savefig api.filter.generate_orthogonal_kernel_1.png
        plt.matshow(kernel);



    See also
    --------
    generate_isotropic_kernel
    generate_kernel
    numpy.bartlett
    numpy.blackman
    numpy.hamming
    numpy.hanning
    numpy.kaiser
    numpy.interp

    """
    kernel = None
    for k1d in kernels:

        # From scalar to 1d
        if np.isscalar(k1d):
            window_func = get_window_func(window_func)
            if isinstance(k1d, int):
                k1d = window_func(k1d)
            else:  # Float case
                size = int(k1d)
                if size != k1d:
                    size += 2
                k1d = np.full(size, fill_value)
                k1d[1:-1] = window_func(size-2)
        else:
            k1d = np.asarray(k1d)

        # Orthogonal merge
        if kernel is None:
            kernel = k1d
        else:
            kernel = np.tensordot(kernel[:, None], k1d[None], axes=1)

    return kernel


def generate_kernel(
        kernel, data, isotropic=False, window_func="ones", fill_value=0.,
        window_args=None, window_kwargs=None):
    """Generate a kernel that is compatible with a given data array

    Parameters
    ----------
    kernel: xarray.DataArray, np.ndarray, int, list, dictorthokernels
        Ready to use kernel or specs to generate it.

        - If an int, the kernel built with ones with a size
          of `kernel` along all dimensions.
        - If a tuple, the kernel is built with ones and a shape
          equal to `kernel`.
        - If a numpy array, it is used as is.

        The final data array is transposed and/or expanded with
        :func:`xoa.coords.transpose` to fit into the input data array.
    data: xarray.DataArray
        Data array that the kernel must be compatible with
    isotropic: bool, tuple
        Tuple of the dimensions over which must be computed isotropically.

    Return
    ------
    xarray.DataArray
        Kernel array with suitable dimensions and shape

    See also
    --------
    generate_isotropic_kernel
    generate_orthogonal_kernel
    xoa.coords.transpose
    """
    # Isotropic
    if isotropic is True:
        isotropic = data.dims

    # Convert to tuple
    if isinstance(kernel, int):
        kernel = (kernel,)*data.ndim

    # Convert tuple to dict with dims
    if isinstance(kernel, tuple):
        if len(kernel) > data.ndim:
            raise XoaError("Too many dimensions for your kernel: {} > {}"
                           .format(len(kernel), data.ndim))
        kernel = dict(item for item in
                      zip(data.dims[-len(kernel):], kernel))

    # Convert list to dict with dims
    elif isinstance(kernel, list):
        kernel = dict((dim, kernel) for dim in data.dims)

    # From an size or 1d kernel for given dimensions
    if isinstance(kernel, dict):

        # Isotropic parameter
        if isotropic:
            if isotropic is True:
                isotropic = data.dims
            elif not set(isotropic).issubset(data.dims):
                raise XoaError("invalid dimensions for isotropic keyword")

        # Split into consistant isotropic and orthogonal kernels
        isokernels_sizes = {}
        isokernel = None
        orthokernels = {}
        for dim, kn in kernel.items():
            if not isotropic or dim not in isotropic:
                orthokernels[dim] = kn
            else:
                if isokernel:
                    if ((np.isscalar(isokernel) and not np.isscalar(kn)) or
                            (not np.isscalar(isokernel) and np.isscalar(kn))):
                        raise XoaError("Inconsistant mix of 1d and scalar "
                                       "kernel specs for building isotropic "
                                       "kernel")
                    if (not np.isscalar(kn) and not np.isscalar(isokernel)
                            and not np.allclose(kn, isokernel)):
                        raise XoaError("Inconsistant 1d kernels for building "
                                       "isotropic kernel")
                else:
                    isokernel = kn
                size = kn if np.isscalar(kn) else len(kn)
                isokernels_sizes[dim] = size

        # Merge orthogonal kernels
        dims = ()
        kernel = None
        if orthokernels:
            dims += tuple(orthokernels.keys())
            window_args = [] if window_args is None else window_args
            window_kwargs = {} if window_kwargs is None else window_kwargs
            window_func = get_window_func(
                window_func, *window_args, **window_kwargs)
            sizes = tuple(orthokernels.values())
            kernel = generate_orthogonal_kernel(
                sizes, window_func=window_func, fill_value=fill_value)

        # Build isotropic kernel
        if isokernel:

            # List/array
            if not np.isscalar(isokernel):
                window_func = get_window_func(isokernel)

            # nD isotropic kernel
            isokernels = generate_isotropic_kernel(
                tuple(isokernels_sizes.values()), window_func,
                fill_value=fill_value)

            # Update final kernel
            dims += tuple(isokernels_sizes.keys())
            if kernel is None:
                kernel = isokernels
            else:
                kernel = np.tensordot(kernel[..., None], isokernels[None, ...])

    # Numpy
    elif isinstance(kernel, np.ndarray):
        if kernel.ndim > data.ndim:
            raise XoaError("too many dimensions for your numpy kernel: {} > {}"
                           .format(kernel.dim, data.ndim))
        dims = data.dims[-kernel.ndim:]

    # Data array
    if not isinstance(kernel, xr.DataArray):
        kernel = xr.DataArray(kernel, dims=dims)
    elif not set(kernel.dims).issubset(set(data.dims)):
        raise XoaError(f"kernel dimensions {kernel.dims} "
                       f"are not a subset of {data.dims}")

    # Finalize
    return xcoords.transpose(kernel, data, mode="insert").astype(data.dtype)


def shapiro_kernel(dims):
    """Generate a shapiro kernel

    Parameters
    ----------
    dims: str, tuple
        Dimension names

    Return
    ------
    xarray.DataArray
        The kernel as a data array with provided dims and a shape of
        ``(3,)*len(dims)``

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.filter import shapiro_kernel
        shapiro_kernel('nx')
        shapiro_kernel(('ny', 'nx'))
        shapiro_kernel(('nt', 'ny', 'nx'))

    """
    if isinstance(dims, str):
        dims = (dims, )
    ndim = len(dims)
    kernel = np.zeros((3,)*ndim, dtype='d')
    indices = np.indices(kernel.shape)
    for idx in indices:
        idx[idx == 2] = 0
        kernel += idx
    return xr.DataArray(kernel, dims=dims)


def _convolve_(data, kernel, normalize):
    """Pure numpy convolution that takes care of nans"""
    import scipy.signal as ss

    # Kernel
    assert data.ndim == kernel.ndim
    if kernel.dtype is not data.dtype:
        xoa_warn(
            "The dtype of your kernel is not the same as that of your data. "
            "Converting it...")
        kernel = kernel.astype(data.dtype)

    # Guess mask
    bad = np.isnan(data)
    with_mask = bad.any()
    if with_mask:
        data = np.where(bad, 0, data)

    # Convolutions
    cdata = ss.convolve(data, kernel, mode='same')
    if normalize or with_mask:
        weights = ss.convolve((~bad).astype('i'), kernel, mode='same')
        weights = np.clip(weights, 0, kernel.sum())

    # Weigthing and masking
    if with_mask:
        bad = np.isclose(weights, 0, atol=float(1e-6*kernel.sum()))
    if normalize:
        if with_mask:
            weights[bad] = 1
        cdata /= weights
    if with_mask:
        cdata[bad.data] = np.nan
    return cdata


def convolve(data, kernel, normalize=False):
    """N-dimensional convolution that takes care of nans

    Parameters
    ----------
    data: xarray.DataArray
        Array to filter
    kernel: int, tuple, numpy.ndarray, xarray.DataArray
        Convolution kernel. See :func:`generate_kernel`.
    normalize: bool
        Divide the convolution product by the local sum weights.
        The result is then a weighted average.

    Return
    ------
    xarray.DataArray
        The filtered array with the same shape, attributes and coordinates
        as the input array.

    See also
    --------
    scipy.signal.convolve
    generate_kernel

    Example
    -------
    .. ipython:: python

        @savefig api.filter.convolve.png
        @suppress
        import xarray as xr, numpy as np, matplotlib.pyplot as plt
        @suppress
        from xoa.filter import convolve
        data = xr.DataArray(np.random.normal(size=(50, 70)), dims=('y', 'x'))
        data[10:20, 10:20] = np.nan # introduce missing data
        kernel = dict(x=[1, 2, 5, 2, 1], y=[1, 2, 1])
        datac = convolve(data, kernel, normalize=True)
        fig, (ax0, ax1) = plt.subplots(ncols=2, figsize=(7, 3))
        kw = dict(vmin=data.min(), vmax=data.max())
        data.plot.pcolormesh(ax=ax0, **kw)
        datac.plot.pcolormesh(ax=ax1, **kw)

    """
    # Adapt the kernel to the data
    kernel = generate_kernel(kernel, data)

    # Numpy convolution
    datac = _convolve_(data.data, kernel.data, normalize)

    # Format
    return xr.DataArray(datac, coords=data.coords, attrs=data.attrs)


def _get_xydims_(data, xdim, ydim):
    if xdim is None:
        xdim = xcoords.get_xdims(data, 'x', allow_positional=False, errors="raise")
    else:
        assert xdim in data.dims, f"Invalid x dimension: {xdim}"
    if ydim is None:
        ydim = xcoords.get_ydims(data, 'y', allow_positional=False, errors="raise")
    else:
        assert ydim in data.dims, f"Invalid y dimension: {ydim}"
    return xdim, ydim


def _reduce_mask_(data, excluded_dims):
    rmask = data.isnull()
    for dim in set(data.dims) - set(excluded_dims):
        rmask = rmask.any(dim=dim)
    return rmask


def _convolve_and_fill_(data, kernel):
    return data.fillna(convolve(data, kernel, normalize=True))


def erode_mask(data, until=1, kernel=None):
    """Erode the horizontal mask using smoothing

    Missing values are filled with the smoothed field in a iterative way.
    Two cases:

        - Erode a fixed number of times.
        - Erode the data mask until there is no missing value where
          a given horirizontal mask is False.

    Parameters
    ----------
    data: xarray.DataArray
        Array of at least 2 dimensions, that are supposed to be horizontal.
    until: xarray.DataArray, int
        Either a minimal mask, or a max number of iteration.
    kernel: None, "shapiro", xarray.DataArray
        Defaults to a :func:`shapiro <shapiro_kernel>` kernel designed
        with all data dimensions.
        If ``kernel`` is provided, it must a compatible with
        :func:`generate_kernel`.

    Return
    ------
    xarray.DataArray
        Data array similar to input array, with its eroded
        along x and y dimensions.

    See also
    --------
    erode_coast
    sharpiro_kernel
    """
    # Kernel
    if kernel is None:
        kernel = "shapiro"
    if isinstance(kernel, str) and kernel == "shapiro":
        kernel = shapiro_kernel(data.dims)
    kernel = generate_kernel(kernel, data)
    kdims = kernel.squeeze().dims

    # Iteration or mask
    if isinstance(until, int):
        niter = until
        mask = None
    else:
        mask = until
        if not set(mask.dims).issubset(data.dims):
            raise XoaError('mask dims must be a subset of data dims')
        mask = xcoords.transpose(mask, data, mode="compat")

    # Filter
    if mask is not None:
        nmask_min = int(mask.sum())
        nmask = np.inf
        while nmask > nmask_min:
            data = _convolve_and_fill_(data, kernel)
            rmask = _reduce_mask_(data, kdims)
            nmask = (mask | rmask).sum()
    else:
        for i in range(niter):
            data = _convolve_and_fill_(data, kernel)

    return data


def erode_coast(data, until=1, kernel=None, xdim=None, ydim=None):
    """Just like :func:`erode_mask` but specialized for the horizontal dimensions

    Parameters
    ----------
    data: xarray.DataArray
        Array of at least 2 dimensions, that are supposed to be horizontal.
    until: array_like, int
        Either a minimal mask, or a max number of iteration.
    kernel:
        Defaults to a :func:`shapiro <shapiro_kernel>` kernel.
        In this case, ``xdim`` and ``ydim`` can be set to the
        horizontal dimensions, otherwise they are inferred.
    xdim: None
        Name of the X dimension, which is infered by default.
    ydim: None
        Name of the Y dimension, which is infered by default.

    Return
    ------
    xarray.DataArray
        Data array similar to input array, with its eroded
        along x and y dimensions.

    See also
    --------
    erode_mask
    sharpiro_kernel
    """
    # We must have X and Y dimensions
    if xdim is None:
        xdim = xcoords.get_xdim(data, errors="raise")
    else:
        assert xdim in data.dims, f"Invalid x dimension: {xdim}"
    if ydim is None:
        ydim = xcoords.get_ydim(data, errors="raise")
    else:
        assert ydim in data.dims, f"Invalid y dimension: {ydim}"

    # Kernel
    if isinstance(kernel, xr.DataArray):
        assert xdim in kernel.dims, f"kernel must have dimension: {xdim}"
        assert ydim in kernel.dims, f"kernel must have dimension: {ydim}"
    elif kernel is None or kernel == "shapiro":
        kernel = shapiro_kernel((ydim, xdim))

    # Mask array
    if not isinstance(until, int):
        assert xdim in until.dims, f"mask must have dimension: {xdim}"
        assert ydim in until.dims, f"mask must have dimension: {ydim}"

    # Filter
    return erode_mask(data, until=until, kernel=kernel)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
xarray and pandas xoa accessors

"""
# Copyright 2020-2021 Shom

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from .misc import ERRORS


class _BasicCFAccessor_(object):
    def __init__(self, obj, cfspecs=None):
        from . import cf

        if cfspecs is None:
            cfspecs = cf.infer_cf_specs(obj)
        self._obj = obj
        self.set_cf_specs(cfspecs)

    def _assign_cf_specs_(self):
        from . import cf

        if self._cfspecs.name:
            self._obj = cf.assign_cf_specs(self._obj, self._cfspecs, register=True)

    def set_cf_specs(self, cfspecs):
        """Set the internal :class:`~xoa.cf.CFSpecs` used by this accessor

        If the specs object has a :cfsec:`registration name <register>`,
        it is assigned to
        the current object and its children with function
        :func:`~xoa.cf.assign_cf_specs`. This set the "cfspecs" encoding
        to the name.

        Parameters
        ----------
        cfspecs: xoa.cf.CFSpecs

        See also
        --------
        :ref:`uses.cf`
        """
        from . import cf

        assert isinstance(cfspecs, cf.CFSpecs)
        self._cfspecs = cfspecs
        self._assign_cf_specs_()

    def get_cf_specs(self):
        """Get the internal :class:`~xoa.cf.CFSpecs` instance used by this accessor

        If not provided at the initialization, it is infered with :func:`xoa.cf.infer_cf_specs`.

        Return
        ------
        xoa.cf.CFSpecs

        See also
        --------
        :ref:`uses.cf`
        """
        return self._cfspecs

    cfspecs = property(fget=get_cf_specs, fset=set_cf_specs, doc="The CFSpecs instance")


class _CFAccessor_(_BasicCFAccessor_):
    _search_category = None

    def __init__(self, obj, cfspecs=None):
        _BasicCFAccessor_.__init__(self, obj, cfspecs)
        self._coords = None
        self._data_vars = None
        self._cache = {}
        self._obj = self.infer_coords()
        self._assign_cf_specs_()

    @ERRORS.format_method_docstring
    def get(self, cf_name, loc="any", single=True, errors="ignore"):
        """Search for a data array or coordinate knowing its generic CF name

        Search is made with the :meth:`~xoa.cf.CFSpecs.search` method of the
        :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Parameters
        ----------
        name: str
        Generic CF name
        loc: str, {{"any", None}}, {{"", False}}
            - str: one of these locations
            - None or "any": any
            - False or "": no location
        {errors}

        See also
        --------
        :meth:`xoa.cf.CFSpecs.search`
        :meth:`xoa.cf.CFCoordSpecs.search`
        :meth:`xoa.cf.CFVarSpecs.search`
        """
        kwargs = dict(cf_name=cf_name, loc=loc, get="obj", single=single, errors=errors)
        if self._search_category is None:
            return self._cfspecs.search(self._obj, **kwargs)
        return self._cfspecs[self._search_category].search(self._obj, **kwargs)

    @ERRORS.format_method_docstring
    def get_coord(self, cf_name, loc="any", single=True, errors="ignore"):
        """Search for a coordinate knowing its generic CF name

        Search is made with the :meth:`~xoa.cf.CFSpecs.search_coord` method of the
        :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Parameters
        ----------
        cf_name: str
            Generic CF name
        loc: str, {{"any", None}}, {{"", False}}
            - str: one of these locations
            - None or "any": any
            - False or "": no location
        {errors}

        See also
        --------
        :meth:`xoa.cf.CFSpecs.search`
        :meth:`xoa.cf.CFCoordSpecs.search`
        :meth:`xoa.cf.CFVarSpecs.search`
        """
        return self._cfspecs.search_coord(
            self._obj, cf_name=cf_name, loc=loc, get="obj", single=single, errors="ignore"
        )

    def __getattr__(self, cf_name):
        """Shortcut to :meth:`get`"""
        return self.get(cf_name, errors="ignore")

    def __getitem__(self, cf_name):
        """Shortcut to :meth:`get`"""
        return self.get(cf_name, errors="ignore")

    def auto_format(self, **kwargs):
        """Rename variables and coordinates and fill their attributes

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Return
        ------
        xarray.Dataset, xarray.DataArray

        See also
        --------
        :meth:`xoa.cf.CFSpecs.auto_format`
        :ref:`uses.cf`
        """
        return self._cfspecs.auto_format(self._obj, **kwargs)

    def fill_attrs(self, **kwargs):
        """Fill missing attributes

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Return
        ------
        xarray.Dataset, xarray.DataArray

        See also
        --------
        :meth:`xoa.cf.CFSpecs.fill_attrs`
        :ref:`uses.cf`
        """
        return self._cfspecs.fill_attrs(self._obj, **kwargs)

    def to_loc(self, **locs):
        """Set the staggered grid location for specified names

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Parameters
        ----------
        locs: dict
            **Keys are root names**, values are new locations.
            A value of `False`, remove the location.
            A value of `None` left it as is.

        See also
        --------
        :meth:`xoa.cf.CFSpecs.to_loc`
        reloc
        """
        return self._cfspecs.to_loc(self._obj, **locs)

    def reloc(self, **locs):
        """Convert given staggered grid locations to other locations

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Parameters
        ----------
        obj: xarray.Dataset, xarray.DataArray
        locs: dict
            **Keys are locations**, values are new locations.
            A value of `False` or `None`, remove the location.

        See also
        --------
        :meth:`xoa.cf.CFSpecs.reloc`
        to_loc
        """
        return self._cfspecs.reloc(self._obj, **locs)

    def infer_coords(self, **kwargs):
        """Infer coordinates and set them as coordinates

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Return
        ------
        xarray.Dataset, xarray.DataArray

        See also
        --------
        :meth:`xoa.cf.CFSpecs.infer_coords`
        :ref:`uses.cf`
        """
        return self.cfspecs.infer_coords(self._obj, **kwargs)

    def decode(self, **kwargs):
        """Rename variables and coordinates to generic names

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.

        Return
        ------
        xarray.Dataset, xarray.DataArray

        See also
        --------
        :meth:`xoa.cf.CFSpecs.decode`
        :meth:`xoa.cf.CFSpecs.encode`
        :ref:`uses.cf`
        """
        return self.cfspecs.decode(self._obj, **kwargs)

    def encode(self, **kwargs):
        """Rename variables and coordinates to specialized names

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.

        Return
        ------
        xarray.Dataset, xarray.DataArray

        See also
        --------
        :meth:`xoa.cf.CFSpecs.encode`
        :meth:`xoa.cf.CFSpecs.decode`
        :ref:`uses.cf`
        """
        return self.cfspecs.encode(self._obj, **kwargs)

    @ERRORS.format_method_docstring
    def get_depth(self, errors="ignore"):
        """Get the depth as computed or recognized by the :meth:`~xoa.cf.CFSpecs`

        If a depth variable cannot be found, it tries to compute either
        from sigma-like coordinates or from layer thinknesses.

        Parameters
        ----------
        {errors}

        Return
        ------
        xarray.DataArray, None

        See also
        --------
        :func:`xoa.coords.get_depth`
        :func:`xoa.grid.decode_cf_dz2depth`
        :func:`xoa.sigma.decode_cf_sigma`
        :ref:`uses.cf`
        """
        from .coords import get_depth

        return get_depth(self._obj, errors=errors)

    @property
    def coords(self):
        """Sub-accessor for coords only"""
        if self._coords is None:
            self._coords = _CoordAccessor_(self._obj, self._cfspecs)
        return self._coords

    @property
    def data_vars(self):
        """Sub-accessor for data_vars only"""
        if self._data_vars is None:
            self._data_vars = _DataVarAccessor__(self._obj, self._cfspecs)
        return self._data_vars


class _CoordAccessor_(_CFAccessor_):
    _search_category = 'coords'

    @property
    def dim(self):
        from .cf import XoaError

        try:
            return self._cfspecs.coords.search_dim(self._obj)[0]
        except XoaError:
            return

    def get_dim(self, cf_arg, loc="any"):
        """Get a dimension name knowing its type or generic CF name

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.

        Parameters
        ----------
        cf_arg: None, {{"x", "y", "z", "t", "f"}}
            Dimension type
        loc:
            Location

        Return
        ------
        str, None
            Dimension name or None of not found

        See also
        --------
        :meth:`xoa.cf.CFSpecs.search_dim`
        """
        cf_arg = cf_arg.lower()
        if not hasattr(self, '_dims'):
            self._dims = {}
            if cf_arg not in self._dims:
                self._dims[cf_arg] = self._cfspecs.coords.search_dim(self._obj, cf_arg, loc=loc)
        return self._dims[cf_arg]

    @property
    def xdim(self):
        """X dimension

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.


        See also
        -------
        :meth:`xoa.cf.CFSpecs.search_dim`
        """
        return self.get_dim("x")

    @property
    def ydim(self):
        """Y dimension name

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.


        See also
        -------
        :meth:`xoa.cf.CFSpecs.search_dim`
        """
        return self.get_dim("y")

    @property
    def zdim(self):
        """Z dimension name

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.


        See also
        -------
        :meth:`xoa.cf.CFSpecs.search_dim`
        """
        return self.get_dim("z")

    @property
    def tdim(self):
        """T (time) dimension name

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.


        See also
        -------
        :meth:`xoa.cf.CFSpecs.search_dim`
        """
        return self.get_dim("t")

    @property
    def fdim(self):
        """F (forecast) dimension name

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.


        See also
        -------
        :meth:`xoa.cf.CFSpecs.search_dim`
        """
        return self.get_dim("f")


class _DataVarAccessor__(_CFAccessor_):
    _search_category = "data_vars"


class CFDatasetAccessor(_CFAccessor_):
    @property
    def ds(self):
        return self._obj


class CFDataArrayAccessor(_CoordAccessor_):
    @property
    def da(self):
        return self._obj

    @property
    def cf_name(self):
        """Get the generic name that matches this array

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.

        See also
        --------
        :meth:`xoa.cf.CFSpecs.match`
        """
        if 'name' not in self._cache:
            category, name = self._cfspecs.match(self._obj)
            self._cache["category"] = category
            self._cache["name"] = name
        return self._cache["name"]

    name = cf_name

    @property
    def attrs(self):
        """Get the generic attributes that matches this array

        This makes use of the :meth:`~xoa.cf.CFSpecs` instance was set at initialisation or
        inferred with :func:`xoa.cf.infer_cf_specs`.
        If no specialized name is declared in the specs, generic names are used.

        See also
        --------
        :meth:`xoa.cf.CFSpecs.get_attrs`
        """
        if "attrs" not in self._cache:
            if self.name:
                cf_attrs = self._cfspecs[self._cache["category"]].get_attrs(
                    self._cache["name"], multi=True
                )
                self._cache["attrs"] = self._cfspecs.sglocator.patch_attrs(
                    self._obj.attrs, cf_attrs
                )
            else:
                self._cache["attrs"] = {}
        return self._cache["attrs"]

    def __getattr__(self, attr):
        if self.name and self.attrs and attr in self.attrs:
            return self._cache["attrs"][attr]
        return _CoordAccessor_.__getattr__(self, attr)


class SigmaAccessor(_BasicCFAccessor_):
    """Dataset accessor to compute depths from sigma-like coordinates

    This follows the CF cnventions.

    Example
    -------
    >>> ds = xr.open_dataset('croco.nc')
    >>> ds = ds.decode_sigma()
    """

    def __init__(self, ds, cfspecs=None):
        assert hasattr(ds, "data_vars"), "ds must be a xarray.Dataset"
        _BasicCFAccessor_.__init__(self, ds, cfspecs)
        self._ds = self._obj

    @ERRORS.format_method_docstring
    def decode(self, rename=False, errors="raise"):
        """Compute depth from sigma coordinates

        Parameters
        ----------
        rename: bool
            Rename and format arrays ot make them compliant with
            :mod:`xoa.cf`
        {errors}

        Return
        ------
        xarray.Dataset

        See also
        --------
        :func:`xoa.sigma.decode_cf_sigma`
        """
        from .sigma import decode_cf_sigma

        return decode_cf_sigma(self._ds, rename=rename, errors=errors)

    def __call__(self):
        """Shortcut to :meth:`decode`"""
        return self.decode()

    def get_sigma_terms(self, loc=None, rename=False):
        """Call :func:`get_sigma_terms` on the dataset


        It operates like this:

        1. Search for the sigma variables.
        2. Parse their ``formula_terms`` attribute.
        3. Create a dict for each locations from names in datasets to
           :mod:`xoa.cf` compliant names that are also used in conversion
           functions.

        Parameters
        ----------
        ds: xarray.Dataset
        loc: str, {"any", None}
            Staggered grid location.
            If any or None, results for all locations are returned.

        Returns
        -------
        dict, dict of dict
            A dict is generated for a given sigma variable,
            whose keys are array names, like ``"sc_r"``,
            and values are :mod:`~xoa.cf` names, like ``"sig"``.
            A special key is the ``"type"`` whose corresponding value
            is the ``standard_name``, stripped from its potential staggered grid
            location indicator.
            If ``loc`` is ``"any"`` or ``None``,
            each dict is embedded in a master dict
            whose keys are staggered grid location. If no location is found,
            the key is set ``None``.

        Raises
        ------
        xoa.sigma.XoaSigmaError
            In case of:

            - inconsistent staggered grid location in dataarrays
              as checked by :meth:`xoa.cf.SGLocator.get_location`
            - no standard_name in sigma/s variable
            - a malformed formula
            - a formula term variable that is not found in the dataset
            - an unknown formula term name

        See also
        --------
        :func:`xoa.sigma.get_sigma_terms`
        """
        from .sigma import get_sigma_terms

        return get_sigma_terms(self._ds, loc=loc, rename=rename)


class XoaDataArrayAccessor(CFDataArrayAccessor):
    @property
    def cf(self):
        """The :class:`CFDataArrayAccessor` subaccessor"""
        if not hasattr(self, "_cf"):
            self._cf = CFDataArrayAccessor(self._ds, self._cfspecs)
        return self._cf


class XoaDatasetAccessor(CFDatasetAccessor):
    @property
    def cf(self):
        """The :class:`~xoa.accessors.CFDatasetAccessor` subaccessor"""
        if not hasattr(self, "_cf"):
            self._cf = CFDatasetAccessor(self._ds, self._cfspecs)
        return self._cf

    @property
    def decode_sigma(self):
        """The :class:`~xoa.accessors.SigmaAccessor` subaccessor for sigma coordinates"""
        if not hasattr(self, "_sigma"):
            self._sigma = SigmaAccessor(self._ds)
        return self._sigma


def _register_xarray_accessors_(dataarrays=None, datasets=None):
    """Silently register xarray accessors"""
    import xarray as xr

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", xr.core.extensions.AccessorRegistrationWarning)
        if dataarrays:
            for name, cls in dataarrays.items():
                xr.register_dataarray_accessor(name)(cls)
        if datasets:
            for name, cls in datasets.items():
                xr.register_dataset_accessor(name)(cls)


def register_cf_accessors(name='xcf'):
    """Register the cf accessors"""
    _register_xarray_accessors_(
        dataarrays={name: CFDataArrayAccessor},
        datasets={name: CFDatasetAccessor},
    )


def register_sigma_accessor(name='decode_sigma'):
    """Register the sigma decoding accessor"""
    _register_xarray_accessors_(datasets={name: SigmaAccessor})


def register_xoa_accessors(name='xoa'):
    """Register the main xoa accessors"""
    _register_xarray_accessors_(
        dataarrays={name: XoaDataArrayAccessor},
        datasets={name: XoaDatasetAccessor},
    )



#!/usr/bin/env python
# coding: utf-8
"""
Compare Hycom3d with a GDP drifter
==================================

In this notebook, we show :

* how to decode dataset so that it is easy to access generic coordinates and variables,
* how to compute depths from layer thicknesses,
* how to interpolate currents from U and V positions to T position on an arakawa C grid,
* how to perform a 4D interpolation with a variable depth coordinate to random positions,
* how to make an horizontal slice of a 4D variable with a variable depth coordinate.

"""
# %%
# Initialisations
# ---------------

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import cartopy.crs as ccrs
import xoa
from xoa.grid import dz2depth, shift
from xoa.regrid import grid2loc, regrid1d
import xoa.cf as xcf
import xoa.geo as xgeo
from xoa.plot import plot_flow

xr.set_options(display_style="text")

# %%
# Register the :ref:`xoa <accessors>` accessors :

xoa.register_accessors()

# %%
# Register the Hycom naming specifications

xcf.register_cf_specs(xoa.get_data_sample("hycom.cfg"))

# %%
# Here is what these CF specifications contain

with open(xoa.get_data_sample("hycom.cfg")) as f:
    print(f.read())

# %%
# Note that most of these specifications are not needed to decode the dataset, i.e to find
# known variables and coordinates, since the default specifications help for that.

# %%
# Read and decode the datasets
# ----------------------------
#
# Hycom velocities
# ~~~~~~~~~~~~~~~~
#
# U and V are stored in separate files at U and V locations on the Hycom Arakawa C-Grid.
# We choose to rename the dataset contents so that we can use generic names to access it.
# We call for that the :meth:`~xarray.Dataset.xoa.decode` accessor method.
# Note that the staggered grid location is automatically appended to horizontal
# coordinates and dimensions of the U and V velocity components since they are
# placed at U and V locations. This prevent any conflict with the bathymetry and
# layer thickness horizontal coordinates and dimensions, which point to the T location.

# %%
# U velocity component
u_hycom = xoa.open_data_sample("hycom.gdp.u.nc").xoa.decode().u.squeeze(drop=True)
print(u_hycom)

# %%
# V velocity component
v_hycom = xoa.open_data_sample("hycom.gdp.v.nc").xoa.decode().v.squeeze(drop=True)

# %%
# Layer thicknesses dataset
hycom = xoa.open_data_sample("hycom.gdp.h.nc").xoa.decode().squeeze(drop=True)
print(hycom)

# %%
# We now compute the depths from the layer thicknesses at the T location
# thanks to the :func:`xoa.grid.dz2depth` function:
# we integrate from a null SSH and interpolate the depth from W locations to T locations.

hycom.coords["depth"] = dz2depth(hycom.dz, centered=True)

# %%
# Then we interpolate the velocity components to the T location with
# the :func:`xoa.grid.shift` function.
# The call to :meth:`reloc <xoa.cf.CFSpecs.reloc>` helps removing the
# staggered grid location prefixes.

ut_hycom = shift(u_hycom, {"x": "left", "y": "left"}).xoa.reloc(u=False)
vt_hycom = shift(v_hycom, {"x": "left", "y": "left"}).xoa.reloc(v=False)
hycom["u"] = ut_hycom.assign_coords(**hycom.coords)
hycom["v"] = vt_hycom.assign_coords(**hycom.coords)

# %%
# So, finally we obtain:

print(hycom)

# %%
# GDP drifter
# ~~~~~~~~~~~
#
# The drifter comes as a `csv` file and we read it as :class:`pandas.DataFrame` instance.

drifter = xoa.open_data_sample("gdp-6203641.csv", header=0, skiprows=[1], parse_dates=[2], index_col=2)

# %%
# Since the sampling is not that nice, we resample it to 3-hour intervals.

drifter = drifter.resample("3H").mean()

# %%
# We convert it to and :class:`xarray.Dataset`, fix the time coordinate and decode it.

drifter = drifter.to_xarray().assign_coords(time=drifter.index.values).xoa.decode()

# %%
# We drop missing values.

drifter = drifter.where(~drifter.lon.isnull() & ~drifter.lat.isnull(), drop=True)

# %%
# We add a constant depth of 15 m.

drifter.coords["depth"] = drifter.lon*0 + 15

# %%
# Here is what we obtain.

print(drifter)

# %%
# Compute and interpolate velocities
# ----------------------------------
#
# We compute the drifter velocity components

drifter["u"] = drifter.lon.differentiate("time", datetime_unit="s") * xgeo.EARTH_RADIUS*np.pi/180
drifter["u"] *= np.cos(np.radians(drifter.lat.values))
drifter["v"] = drifter.lat.differentiate("time", datetime_unit="s") * xgeo.EARTH_RADIUS*np.pi/180

# %%
# We make a 4D linear interpolation the Hycom velocity to the drifter positions with :func:`xoa.regrid.grid2loc`.

uloc = grid2loc(hycom["u"], drifter)
vloc = grid2loc(hycom["v"], drifter)

# %%
# Instead of just showing the velocities along the drifter positions,
# we can plot the mean model velocities over the same period as a background.
# So we interpolate them at 15 m with :func:`xoa.regrid.regrid1d` and compute a time average.

d15 = xr.DataArray([15.], dims="depth", name="depth")
uh15 = regrid1d(hycom["u"], d15).squeeze(drop=True).mean(dim="time")
vh15 = regrid1d(hycom["v"], d15).squeeze(drop=True).mean(dim="time")

# %%
# Plot
# ----
#
# The geographic extent is easily computed with the :func:`xoa.geo.get_extent` function.
# The background currents are plotted with the :func:`xoa.plot.plot_flow` function.

pmerc = ccrs.Mercator()
pcarr = ccrs.PlateCarree()
fig, ax = plt.subplots(figsize=(7, 7), subplot_kw={"facecolor": "teal", "projection": pmerc})
ax.gridlines(draw_labels=True, dms=True)
ax.set_extent(xgeo.get_extent(uh15))
kwqv = dict(scale_units="dots", scale=0.1/20, units="dots", transform=pcarr)
plot_flow(uh15, vh15, axes=ax, transform=pcarr, color="w", alpha=(.3, 1), linewidth=.6)
qv = ax.quiver(uloc.lon.values, uloc.lat.values, uloc.values, vloc.values,
    color="w", width=2, label="Model", **kwqv)
ax.plot(drifter.lon.values, drifter.lat.values, '-', color="C1", transform=pcarr, lw=.5)
ax.quiver(drifter.lon.values, drifter.lat.values, drifter.u.values, drifter.v.values,
    color="C1", label="Drifter", width=2, **kwqv)
plt.quiverkey(qv, 0.1, 1.06, 0.1, r"0.1 $m\,s^{-1}$", color="k", alpha=1, labelpos="E")
plt.legend();

# %%
# The discrepancies between the lagrangian and mean eulerian currents highlight
# the variability due to the mesocale activity.
# The differences between the observed and modeled currents are partly due to
# the lack of data assimilation in the model.



# -*- coding: utf-8 -*-
"""
Colors and colormaps utilities
"""
# Copyright 2020-2021 Shom
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmocean
import matplotlib.pyplot as plt

from .__init__ import xoa_warn


def crop_cmap(cmapin, vmin, vmax, pivot=0):
    """Crop a colormap so that it is centered around pivot

    This is a wrapper for :func:`cmocean.tools.crop`.

    Parameters
    ----------
    cmap: colormap
        Compatible with :func:`matplotlib.pyplot.get_cmap`.
    vmin: float
        Min data value
    vmax: float
        Max data value
    pivot: float
        The colormap will be centered on this value.
        Should be lower than vmax et greater than vmin.
    """
    cmapin = plt.get_cmap(cmapin)
    return cmocean.tools.crop(cmapin, vmin, vmax, pivot)


class CmapAdapter(object):
    """Adapt a given colormap to data

    Parameters
    ----------
    cmap:
        The colormap to adapt
    specs: str, None
        Transformation specifications or None.

        When a string, it must have the format ``"<type><value>"``, where
        type is ``"piv"`` or ``"cyc"``, and value is convertible to a float.

        If type is ``"piv"``, that colormap is expected typically
        to be **diverging**, and is cropped
        using :func:`crop_cmap`, after min and max value are set with
        :meth:`set_vlim`.

        If type is ``"cyc"``, min is set to 0 and max to ``value``,
        and the colormap is
        expected to be **circular**, like ``"cmo.phase"``.

    Example
    -------
    .. ipython:: python

        @suppress
        from xoa.color import CmapAdapter
        @suppress
        import matplotlib.pyplot as plt, numpy as np
        cma = CmapAdapter('cmo.balance', 'piv0')
        data = np.arange(100).reshape(10, 10) - 20
        cma.set_vlim(data.min(), data.max())
        plt.contourf(data, **cma.get_dict());
        @savefig api.color.cmapadapter.png
        plt.colorbar();
    """

    def __init__(self, cmap, specs=None):
        self.cmap = plt.get_cmap(cmap)
        self.vmin = self.vmax = None
        if specs is None:
            self.specs = None
        if specs.startswith("piv"):
            self.specs = ("pivot", float(specs[3:]))
        elif specs.startswith("cyc"):
            self.specs = ("cycle", float(specs[3:]))
            self.vmin = 0
            self.vmax = self.specs[1]

    def set_vlim(self, vmin, vmax):
        """Set the min and max data value for scaling"""
        if self.specs and self.specs[0] == "cycle":
            vmin, vmax = 0, 360.0
        self.vmin = vmin
        self.vmax = vmax
        return vmin, vmax

    def get_cmap(self):
        """Get the adapted colormap"""
        if self.specs[0] == "pivot":
            if self.vmin is None or self.vmax is None:
                xoa_warn(
                    "cmap not adapted since vmin and vmin are not set"
                )
                return self.cmap
            return crop_cmap(self.cmap, self.vmin, self.vmax, self.specs[1])

    def get_dict(self):
        """The specs for plots as a dict whose keys are cmap, vmin and vmax"""
        return dict(cmap=self.get_cmap(), vmin=self.vmin, vmax=self.vmax)
